/*----- PROTECTED REGION ID(OceanOptics.cpp) ENABLED START -----*/
static const char *RcsId = "$Id: OceanOptics.cpp,v 1.19 2019-04-23 13:32:28 graziano Exp $";
//=============================================================================
//
// file :        OceanOptics.cpp
//
// description : C++ source for the OceanOptics class and its commands.
//               The class is derived from Device. It represents the
//               CORBA servant object which will be accessed from the
//               network. All commands which can be executed on the
//               OceanOptics are implemented in this file.
//
// project :     SolarLaserSystems S100
//
// This file is part of Tango device class.
// 
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
// 
// $Author: graziano $
//
// $Revision: 1.19 $
// $Date: 2019-04-23 13:32:28 $
//
// $HeadURL:  $
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include <tango.h>
#include <OceanOptics.h>
#include <OceanOpticsClass.h>
#include "readthread.h"
#include <arpa/inet.h>
//#include "check_get_rt_buffer.h"
#include "tcheck_get_rt_buffer.hpp"

/*----- PROTECTED REGION END -----*/	//	OceanOptics.cpp

/**
 *  OceanOptics class description:
 *    Tango device server for OceanOptics compact wide-range spectrometer
 */

//================================================================
//  The following table gives the correspondence
//  between command and method names.
//
//  Command name       |  Method name
//================================================================
//  State              |  Inherited (no method)
//  Status             |  Inherited (no method)
//  GetVersion         |  get_version
//  Stop               |  stop
//  StartSingle        |  start_single
//  StartContinuous    |  start_continuous
//  GetLastData        |  get_last_data
//  AcquireNoise       |  acquire_noise
//  SaveReferenceData  |  save_reference_data
//  OpenShutter        |  open_shutter
//  CloseShutter       |  close_shutter
//  GetData            |  get_data
//  GetGaussFitCenter  |  get_gauss_fit_center
//  GetGaussFitWidth   |  get_gauss_fit_width
//  GetFWHMCenter      |  get_fwhmcenter
//  GetFWHMWidth       |  get_fwhmwidth
//  ExportWavelength   |  export_wavelength
//  GetArea            |  get_area
//  GetWave            |  get_wave
//  GetLambda          |  get_lambda
//================================================================

//================================================================
//  Attributes managed are:
//================================================================
//  ExposureTime                |  Tango::DevDouble	Scalar
//  IncrSensitivity             |  Tango::DevBoolean	Scalar
//  SyncMode                    |  Tango::DevLong	Scalar
//  NoiseSubtraction            |  Tango::DevBoolean	Scalar
//  GaussFitEnable              |  Tango::DevBoolean	Scalar
//  FWHMEnable                  |  Tango::DevBoolean	Scalar
//  GaussFitCenter              |  Tango::DevDouble	Scalar
//  GaussFitWidth               |  Tango::DevDouble	Scalar
//  FWHMCenter                  |  Tango::DevDouble	Scalar
//  FWHMWidth                   |  Tango::DevDouble	Scalar
//  Center                      |  Tango::DevDouble	Scalar
//  Width                       |  Tango::DevDouble	Scalar
//  FitType                     |  Tango::DevLong	Scalar
//  FWHMAvg                     |  Tango::DevLong	Scalar
//  NoiseLevel                  |  Tango::DevLong	Scalar
//  DataReferenceFile           |  Tango::DevString	Scalar
//  AutoExposure                |  Tango::DevBoolean	Scalar
//  LightSourceIntensity        |  Tango::DevDouble	Scalar
//  RtMode                      |  Tango::DevBoolean	Scalar
//  LightSourceEnable           |  Tango::DevBoolean	Scalar
//  IntensityCalibrationEnable  |  Tango::DevBoolean	Scalar
//  Area                        |  Tango::DevDouble	Scalar
//  AreaRegion                  |  Tango::DevDouble	Scalar
//  AreaEnable                  |  Tango::DevBoolean	Scalar
//  WidhtCorrectionEnable       |  Tango::DevBoolean	Scalar
//  NumPixels                   |  Tango::DevLong	Spectrum  ( max = 2)
//  Data                        |  Tango::DevDouble	Spectrum  ( max = 5000)
//  RegionOfInterest            |  Tango::DevDouble	Spectrum  ( max = 2)
//  AnalysisRegion              |  Tango::DevDouble	Spectrum  ( max = 2)
//  GaussFit                    |  Tango::DevDouble	Spectrum  ( max = 4)
//  FWHM                        |  Tango::DevDouble	Spectrum  ( max = 3)
//  GaussReference              |  Tango::DevDouble	Spectrum  ( max = 4)
//  DataReference               |  Tango::DevDouble	Spectrum  ( max = 5000)
//  IntensityCalibration        |  Tango::DevDouble	Spectrum  ( max = 10000)
//  NoiseProfile                |  Tango::DevDouble	Spectrum  ( max = 5000)
//  Wave                        |  Tango::DevDouble	Spectrum  ( max = 5000)
//  Lambda                      |  Tango::DevDouble	Spectrum  ( max = 5000)
//================================================================

namespace OceanOptics_ns
{
/*----- PROTECTED REGION ID(OceanOptics::namespace_starting) ENABLED START -----*/

//	static initializations

/*----- PROTECTED REGION END -----*/	//	OceanOptics::namespace_starting

//--------------------------------------------------------
/**
 *	Method      : OceanOptics::OceanOptics()
 *	Description : Constructors for a Tango device
 *                implementing the classOceanOptics
 */
//--------------------------------------------------------
OceanOptics::OceanOptics(Tango::DeviceClass *cl, string &s)
 : TANGO_BASE_CLASS(cl, s.c_str())
{
	/*----- PROTECTED REGION ID(OceanOptics::constructor_1) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::constructor_1
}
//--------------------------------------------------------
OceanOptics::OceanOptics(Tango::DeviceClass *cl, const char *s)
 : TANGO_BASE_CLASS(cl, s)
{
	/*----- PROTECTED REGION ID(OceanOptics::constructor_2) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::constructor_2
}
//--------------------------------------------------------
OceanOptics::OceanOptics(Tango::DeviceClass *cl, const char *s, const char *d)
 : TANGO_BASE_CLASS(cl, s, d)
{
	/*----- PROTECTED REGION ID(OceanOptics::constructor_3) ENABLED START -----*/
	init_device();
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::constructor_3
}

//--------------------------------------------------------
/**
 *	Method      : OceanOptics::delete_device()
 *	Description : will be called at device destruction or at init command
 */
//--------------------------------------------------------
void OceanOptics::delete_device()
{
	DEBUG_STREAM << "OceanOptics::delete_device() " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::delete_device) ENABLED START -----*/
	
	//	Delete device allocated objects
	abortflag = true;

	Tango::DbDatum gr("DefaultGaussRef");
	vector<double> gref;
	gref.push_back(attr_GaussReference_read[0]);
	gref.push_back(attr_GaussReference_read[1]);
	gref.push_back(attr_GaussReference_read[2]);
	gref.push_back(attr_GaussReference_read[3]);
	gr << gref;
	Tango::DbDatum ar("AnalysisRegion");
	vector<double> anre;
	anre.push_back(attr_AnalysisRegion_read[0]);
	anre.push_back(attr_AnalysisRegion_read[1]);
	ar << anre;
	Tango::DbDatum ri("RegionOfInterest");
	vector<double> rein;
	rein.push_back(attr_RegionOfInterest_read[0]);
	rein.push_back(attr_RegionOfInterest_read[1]);
	ri << rein;
	Tango::DbDatum et("ExposureTime");
	et << *attr_ExposureTime_read;
	Tango::DbDatum sm("SyncMode");
	sm << *attr_SyncMode_read;
	Tango::DbData data;
	data.push_back(gr);
	data.push_back(ar);
	data.push_back(ri);
	data.push_back(et);
	data.push_back(sm);
	Tango::Database *db = new Tango::Database();
	try
	{
		db->set_timeout_millis(15000);
		db->put_device_property(get_name(), data);
	}
	catch(Tango::DevFailed &e)
	{
		stringstream o;
		o << " Error saving properties='" << e.errors[0].desc << "'";
		WARN_STREAM << __FUNCTION__<< o.str();
	}
	delete db;

/*	Tango::DbDatum dsr("DefaultStartRegion");
	dsr << attr_RegionOfInterest_read[0];
	Tango::DbData data2;
	data2.push_back(dsr);
	get_db_device()->put_property(data2);

	Tango::DbDatum der("DefaultEndRegion");
	der << attr_RegionOfInterest_read[1];
	Tango::DbData data3;
	data3.push_back(der);
	get_db_device()->put_property(data3);*/

	usleep(500000);
	//	Delete device's allocated object
	delete attr_ExposureTime_read;
	delete attr_IncrSensitivity_read;
	delete attr_SyncMode_read;
	delete [] attr_NumPixels_read;
	delete [] attr_RegionOfInterest_read;
	delete [] attr_AnalysisRegion_read;
	delete attr_NoiseSubtraction_read;
	delete attr_GaussFitEnable_read;
	delete attr_AreaEnable_read;
	delete attr_FWHMEnable_read;
	delete [] attr_GaussFit_read;
	delete attr_GaussFitCenter_read;
	delete attr_GaussFitWidth_read;
	delete [] attr_FWHM_read;
	delete attr_FWHMCenter_read;
	delete attr_FWHMWidth_read;
	delete attr_Center_read;
	delete attr_Width_read;
	delete [] attr_GaussReference_read;
	//delete [] attr_DataReference_read;
	//delete [] attr_Data_read;
	//delete [] attr_Wave_read;
	delete [] x_values;
	if(pLiveData_avg)
	{
		delete [] pLiveData_avg;
		pLiveData_avg = NULL;
	}
	if(pLiveData)
	{
		pLiveData = NULL;
		delete [] pLiveData;
	}
	if(pNoiseMask_avg)
	{
		delete [] pNoiseMask_avg;
		pNoiseMask_avg = NULL;
	}
	if(pReferenceData)
	{
		delete [] pReferenceData;
		pReferenceData = NULL;
	}
	/*if(pMeasData)
	{
		pMeasData = NULL;
		delete [] pMeasData;
	}*/
	if(pFrameBuffer)
	{
		pFrameBuffer = NULL;
		delete [] pFrameBuffer;
	}
	delete m_gaussfit;
	delete m_asyngaussfit;
	delete m_confiteorfit;

	for(int i=0; i<NUM_BUF; i++)
	{
		tbuf[i].delete_rt_buffer();
	}
	tbuf_array.delete_rt_buffer();
	tbuf_array_w.delete_rt_buffer();
	tbuf_array_l.delete_rt_buffer();

	disconnect_device();

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::delete_device
	delete[] attr_LightSourceIntensity_read;
	delete[] attr_RtMode_read;
	delete[] attr_LightSourceEnable_read;
	delete[] attr_IntensityCalibrationEnable_read;
	delete[] attr_Area_read;
	delete[] attr_AreaRegion_read;
	delete[] attr_WidhtCorrectionEnable_read;
	delete[] attr_IntensityCalibration_read;
}

//--------------------------------------------------------
/**
 *	Method      : OceanOptics::init_device()
 *	Description : will be called at device initialization.
 */
//--------------------------------------------------------
void OceanOptics::init_device()
{
	DEBUG_STREAM << "OceanOptics::init_device() create device " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::init_device_before) ENABLED START -----*/
	
	//	Initialization before get_device_property() call

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::init_device_before
	

	//	Get the device properties from database
	get_device_property();
	
	attr_LightSourceIntensity_read = new Tango::DevDouble[1];
	attr_RtMode_read = new Tango::DevBoolean[1];
	attr_LightSourceEnable_read = new Tango::DevBoolean[1];
	attr_IntensityCalibrationEnable_read = new Tango::DevBoolean[1];
	attr_Area_read = new Tango::DevDouble[1];
	attr_AreaRegion_read = new Tango::DevDouble[1];
	attr_WidhtCorrectionEnable_read = new Tango::DevBoolean[1];
	attr_IntensityCalibration_read = new Tango::DevDouble[10000];
	/*----- PROTECTED REGION ID(OceanOptics::init_device) ENABLED START -----*/
	defaultStartRegion = floor(defaultStartRegion*1000.0)*0.001;
	defaultEndRegion = floor(defaultEndRegion*1000.0)*0.001;
	for(int i=0; i<10000; i++)
		attr_IntensityCalibration_read[i] = 1.0;
	*attr_IntensityCalibrationEnable_read = false;
	*attr_LightSourceEnable_read = false;
	*attr_LightSourceIntensity_read = 0;
	//	Initialize device

	attr_ExposureTime_read = new Tango::DevDouble();
	attr_IncrSensitivity_read = new Tango::DevBoolean();
	attr_SyncMode_read = new Tango::DevLong();
	attr_NumPixels_read = new Tango::DevLong[2];
	attr_RegionOfInterest_read = new Tango::DevDouble[2];
	attr_AnalysisRegion_read = new Tango::DevDouble[2];
	attr_NoiseSubtraction_read = new Tango::DevBoolean();
	attr_GaussFitEnable_read = new Tango::DevBoolean();
	attr_AreaEnable_read = new Tango::DevBoolean();
	attr_FWHMEnable_read = new Tango::DevBoolean();
	attr_GaussFit_read = new Tango::DevDouble[4];
	attr_GaussFitCenter_read = new Tango::DevDouble();
	attr_GaussFitWidth_read = new Tango::DevDouble();
	attr_FWHM_read = new Tango::DevDouble[3];
	attr_FWHMCenter_read = new Tango::DevDouble();
	attr_FWHMWidth_read = new Tango::DevDouble();
	attr_Center_read = new Tango::DevDouble();
	attr_Width_read = new Tango::DevDouble();
	attr_GaussReference_read = new Tango::DevDouble[4];
	attr_FitType_read = &attr_FitType_write;
	attr_FWHMAvg_read = &attr_FWHMAvg_write;
	attr_NoiseLevel_read = &attr_NoiseLevel_write;
	//attr_DataReference_read = new Tango::DevLong[...];
	attr_AutoExposure_read = &attr_AutoExposure_write;

	//noise_level = *attr_NoiseLevel_read;
	attr_DataReference_read = NULL;
	attr_Data_read = NULL;
	attr_Wave_read = NULL;
	*attr_NoiseSubtraction_read = false;
	*attr_GaussFitEnable_read = false;
	*attr_AreaEnable_read = false;
	*attr_FWHMEnable_read = false;
	*attr_WidhtCorrectionEnable_read = false;
	//data_counter_live = 0;
	//data_counter_meas = 0;
	FrameBufferIndex = 0;
	//pMeasData = NULL;
	pLiveData = NULL;
	pReferenceData = NULL;
	pFrameBuffer = NULL;

	*attr_ExposureTime_read = exposureTime;
	*attr_SyncMode_read = syncMode;

	attr_AnalysisRegion_read[0] = attr_RegionOfInterest_read[0] = defaultStartRegion;
	attr_AnalysisRegion_read[1] = attr_RegionOfInterest_read[1] = defaultEndRegion;
	if(regionOfInterest.size() >= 2)
	{
		attr_RegionOfInterest_read[0] = regionOfInterest[0];
		attr_RegionOfInterest_read[1] = regionOfInterest[1];
	}
	if(analysisRegion.size() >= 2)
	{
		attr_AnalysisRegion_read[0] = analysisRegion[0];
		attr_AnalysisRegion_read[1] = analysisRegion[1];
	}
	if(defaultGaussRef.size() == 4)
	{
		int i=0;
		for(vector<double>::iterator it = defaultGaussRef.begin(); it != defaultGaussRef.end(); it++)
			attr_GaussReference_read[i++] = *it;
		/*attr_GaussReference_read[0] = defaultGaussRef[0];
		attr_GaussReference_read[1] = defaultGaussRef[1];
		attr_GaussReference_read[2] = defaultGaussRef[2];
		attr_GaussReference_read[3] = defaultGaussRef[3];*/
	}
	else
		attr_GaussReference_read[0] = attr_GaussReference_read[1] = attr_GaussReference_read[2] = attr_GaussReference_read[3]= 0;

	ostringstream tempstr;
	tempstr << dataReferenceFile << ends;
	strcpy(attr_strchar,tempstr.str().c_str());

	abortflag = false;
	live_mode = false;
	init_live_mode = false;
	measure_mode = false;
	init_measure_mode = false;
	acquire_noise_mode = false;
	init_acquire_noise_mode = false;
	noise_level = 0;
	NumPixel = 0;

	y_max_mean = 0;
	memset(exposures, 0, sizeof(int)*AUTO_EXPOSURE_MEAN);
	exposure_idx = 0;

//	m_gaussfit = new GaussFit(NUM_GAUSS_PARAMETER, gauss_f, gauss_df, gauss_fdf);
//	m_gaussfit = new Fit(NUM_GAUSS_PARAMETER, gauss_f, gauss_df, gauss_fdf);
	m_gaussfit = new Fit(GAUSS_FIT, GAUSS_FIT_NUM_PARAM, 1);
	m_asyngaussfit = new Fit(GAUSS_FIT_ASYMM, GAUSS_FIT_ASYMM_NUM_PARAM, 1);
	m_confiteorfit = new Fit(CONFITEOR2_FIT, CONFITEOR_FIT_NUM_PARAM, 1);
	loop = new readthread(this);

	x_values = NULL;
	pLiveData_avg = NULL;
	pNoiseMask_avg = NULL;
	pFrameBuffer = NULL;

	for(int i=0; i<NUM_BUF; i++)
	{
		tbuf[i].init_rt_buffer(BUF_SIZE);
		tbuf[i].set_mode(tbuffer::BUF_RT_CIRCULAR);
		//set_fill_hole(BUF_RT_FILL_0, buf[i]);
		//set_fill_hole(BUF_RT_NO_FILL, buf[i]);
		tbuf[i].set_fill_hole(BUF_RT_FILL_1);
	}
	tbuf_array.init_rt_buffer(BUF_SIZE);
	tbuf_array.set_mode(tbuffer::BUF_RT_CIRCULAR);
	tbuf_array.set_fill_hole(BUF_RT_FILL_1);

	tbuf_array_w.init_rt_buffer(BUF_SIZE);
	tbuf_array_w.set_mode(tbuffer::BUF_RT_CIRCULAR);
	tbuf_array_w.set_fill_hole(BUF_RT_FILL_1);

	tbuf_array_l.init_rt_buffer(BUF_SIZE);
	tbuf_array_l.set_mode(tbuffer::BUF_RT_CIRCULAR);
	tbuf_array_l.set_fill_hole(BUF_RT_FILL_1);

	connect_device(false);

	if(intensityCalibrationFile.length() > 0)
		load_intensity_cal_file(intensityCalibrationFile);

	loop->start();
	if(autoStart && device_id != -1)
	{
		try
		{
			start_continuous();
		}
		catch(Tango::DevFailed &e)
		{
			INFO_STREAM << __func__ << ": error autostarting=" << e.errors[0].desc;
		}
	}
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::init_device
}

//--------------------------------------------------------
/**
 *	Method      : OceanOptics::get_device_property()
 *	Description : Read database to initialize property data members.
 */
//--------------------------------------------------------
void OceanOptics::get_device_property()
{
	/*----- PROTECTED REGION ID(OceanOptics::get_device_property_before) ENABLED START -----*/

	//	Initialize property data members


	/*----- PROTECTED REGION END -----*/	//	OceanOptics::get_device_property_before


	//	Read device properties from database.
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("DeviceId"));
	dev_prop.push_back(Tango::DbDatum("SerialNumber"));
	dev_prop.push_back(Tango::DbDatum("NFramesBuffered"));
	dev_prop.push_back(Tango::DbDatum("DefaultStartRegion"));
	dev_prop.push_back(Tango::DbDatum("DefaultEndRegion"));
	dev_prop.push_back(Tango::DbDatum("SleepReading"));
	dev_prop.push_back(Tango::DbDatum("DefaultGaussRef"));
	dev_prop.push_back(Tango::DbDatum("DataReferenceFile"));
	dev_prop.push_back(Tango::DbDatum("PreferFWHM"));
	dev_prop.push_back(Tango::DbDatum("AutoExposureRange"));
	dev_prop.push_back(Tango::DbDatum("AutoExposureStep"));
	dev_prop.push_back(Tango::DbDatum("AutoExposureAmpRange"));
	dev_prop.push_back(Tango::DbDatum("IntensityCalibrationFile"));
	dev_prop.push_back(Tango::DbDatum("NoiseProfileFile"));
	dev_prop.push_back(Tango::DbDatum("WavelengthAttr"));
	dev_prop.push_back(Tango::DbDatum("AreaInAnalysisRegion"));
	dev_prop.push_back(Tango::DbDatum("WidthCorrection"));
	dev_prop.push_back(Tango::DbDatum("AnalysisRegion"));
	dev_prop.push_back(Tango::DbDatum("RegionOfInterest"));
	dev_prop.push_back(Tango::DbDatum("AutoStart"));
	dev_prop.push_back(Tango::DbDatum("ExposureTime"));
	dev_prop.push_back(Tango::DbDatum("SyncMode"));

	//	is there at least one property to be read ?
	if (dev_prop.size()>0)
	{
		//	Call database and extract values
		if (Tango::Util::instance()->_UseDb==true)
			get_db_device()->get_property(dev_prop);
	
		//	get instance on OceanOpticsClass to get class property
		Tango::DbDatum	def_prop, cl_prop;
		OceanOpticsClass	*ds_class =
			(static_cast<OceanOpticsClass *>(get_device_class()));
		int	i = -1;

		//	Try to initialize DeviceId from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  deviceId;
		else {
			//	Try to initialize DeviceId from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  deviceId;
		}
		//	And try to extract DeviceId value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  deviceId;

		//	Try to initialize SerialNumber from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  serialNumber;
		else {
			//	Try to initialize SerialNumber from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  serialNumber;
		}
		//	And try to extract SerialNumber value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  serialNumber;

		//	Try to initialize NFramesBuffered from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  nFramesBuffered;
		else {
			//	Try to initialize NFramesBuffered from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  nFramesBuffered;
		}
		//	And try to extract NFramesBuffered value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  nFramesBuffered;

		//	Try to initialize DefaultStartRegion from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  defaultStartRegion;
		else {
			//	Try to initialize DefaultStartRegion from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  defaultStartRegion;
		}
		//	And try to extract DefaultStartRegion value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  defaultStartRegion;

		//	Try to initialize DefaultEndRegion from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  defaultEndRegion;
		else {
			//	Try to initialize DefaultEndRegion from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  defaultEndRegion;
		}
		//	And try to extract DefaultEndRegion value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  defaultEndRegion;

		//	Try to initialize SleepReading from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  sleepReading;
		else {
			//	Try to initialize SleepReading from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  sleepReading;
		}
		//	And try to extract SleepReading value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  sleepReading;

		//	Try to initialize DefaultGaussRef from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  defaultGaussRef;
		else {
			//	Try to initialize DefaultGaussRef from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  defaultGaussRef;
		}
		//	And try to extract DefaultGaussRef value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  defaultGaussRef;

		//	Try to initialize DataReferenceFile from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  dataReferenceFile;
		else {
			//	Try to initialize DataReferenceFile from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  dataReferenceFile;
		}
		//	And try to extract DataReferenceFile value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  dataReferenceFile;

		//	Try to initialize PreferFWHM from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  preferFWHM;
		else {
			//	Try to initialize PreferFWHM from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  preferFWHM;
		}
		//	And try to extract PreferFWHM value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  preferFWHM;

		//	Try to initialize AutoExposureRange from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  autoExposureRange;
		else {
			//	Try to initialize AutoExposureRange from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  autoExposureRange;
		}
		//	And try to extract AutoExposureRange value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  autoExposureRange;

		//	Try to initialize AutoExposureStep from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  autoExposureStep;
		else {
			//	Try to initialize AutoExposureStep from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  autoExposureStep;
		}
		//	And try to extract AutoExposureStep value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  autoExposureStep;

		//	Try to initialize AutoExposureAmpRange from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  autoExposureAmpRange;
		else {
			//	Try to initialize AutoExposureAmpRange from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  autoExposureAmpRange;
		}
		//	And try to extract AutoExposureAmpRange value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  autoExposureAmpRange;

		//	Try to initialize IntensityCalibrationFile from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  intensityCalibrationFile;
		else {
			//	Try to initialize IntensityCalibrationFile from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  intensityCalibrationFile;
		}
		//	And try to extract IntensityCalibrationFile value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  intensityCalibrationFile;

		//	Try to initialize NoiseProfileFile from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  noiseProfileFile;
		else {
			//	Try to initialize NoiseProfileFile from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  noiseProfileFile;
		}
		//	And try to extract NoiseProfileFile value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  noiseProfileFile;

		//	Try to initialize WavelengthAttr from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  wavelengthAttr;
		else {
			//	Try to initialize WavelengthAttr from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  wavelengthAttr;
		}
		//	And try to extract WavelengthAttr value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  wavelengthAttr;

		//	Try to initialize AreaInAnalysisRegion from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  areaInAnalysisRegion;
		else {
			//	Try to initialize AreaInAnalysisRegion from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  areaInAnalysisRegion;
		}
		//	And try to extract AreaInAnalysisRegion value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  areaInAnalysisRegion;

		//	Try to initialize WidthCorrection from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  widthCorrection;
		else {
			//	Try to initialize WidthCorrection from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  widthCorrection;
		}
		//	And try to extract WidthCorrection value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  widthCorrection;

		//	Try to initialize AnalysisRegion from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  analysisRegion;
		else {
			//	Try to initialize AnalysisRegion from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  analysisRegion;
		}
		//	And try to extract AnalysisRegion value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  analysisRegion;

		//	Try to initialize RegionOfInterest from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  regionOfInterest;
		else {
			//	Try to initialize RegionOfInterest from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  regionOfInterest;
		}
		//	And try to extract RegionOfInterest value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  regionOfInterest;

		//	Try to initialize AutoStart from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  autoStart;
		else {
			//	Try to initialize AutoStart from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  autoStart;
		}
		//	And try to extract AutoStart value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  autoStart;

		//	Try to initialize ExposureTime from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  exposureTime;
		else {
			//	Try to initialize ExposureTime from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  exposureTime;
		}
		//	And try to extract ExposureTime value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  exposureTime;

		//	Try to initialize SyncMode from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  syncMode;
		else {
			//	Try to initialize SyncMode from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  syncMode;
		}
		//	And try to extract SyncMode value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  syncMode;

	}

	/*----- PROTECTED REGION ID(OceanOptics::get_device_property_after) ENABLED START -----*/

	//	Check device property data members init

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::get_device_property_after
}

//--------------------------------------------------------
/**
 *	Method      : OceanOptics::always_executed_hook()
 *	Description : method always executed before any command is executed
 */
//--------------------------------------------------------
void OceanOptics::always_executed_hook()
{
	DEBUG_STREAM << "OceanOptics::always_executed_hook()  " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::always_executed_hook) ENABLED START -----*/

	//	code always executed before all requests
	timespec now;
	clock_gettime(CLOCK_MONOTONIC,&now);
	double diff_time = (now.tv_sec - last_read.tv_sec) + (double)(now.tv_nsec - last_read.tv_nsec)/1000000000;
	if(live_mode && device_id != -1 && diff_time>2)
	{
		set_state(Tango::ALARM);
		set_status("Trigger Missing");
	}
	else if(live_mode && device_id != -1 && diff_time<=2 && get_state()==Tango::ALARM)
	{
		set_state(Tango::RUNNING);
		set_status("Continuous Mode");
	}

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::always_executed_hook
}

//--------------------------------------------------------
/**
 *	Method      : OceanOptics::read_attr_hardware()
 *	Description : Hardware acquisition for attributes
 */
//--------------------------------------------------------
void OceanOptics::read_attr_hardware(TANGO_UNUSED(vector<long> &attr_list))
{
	DEBUG_STREAM << "OceanOptics::read_attr_hardware(vector<long> &attr_list) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_attr_hardware) ENABLED START -----*/

	//	Add your own code
	//DEBUG_STREAM << "OceanOptics::read_attr_hardware(vector<long> &attr_list) entering... "<< endl;
	//	Add your own code here

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_attr_hardware
}
//--------------------------------------------------------
/**
 *	Method      : OceanOptics::write_attr_hardware()
 *	Description : Hardware writing for attributes
 */
//--------------------------------------------------------
void OceanOptics::write_attr_hardware(TANGO_UNUSED(vector<long> &attr_list))
{
	DEBUG_STREAM << "OceanOptics::write_attr_hardware(vector<long> &attr_list) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::write_attr_hardware) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_attr_hardware
}

//--------------------------------------------------------
/**
 *	Read attribute ExposureTime related method
 *	Description: Exposure time in milliseconds
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_ExposureTime(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_ExposureTime(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_ExposureTime) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_ExposureTime(Tango::Attribute &attr) entering... "<< endl;
	//if(!m_usbCCD.GetParameters())
	//	INFO_STREAM << "OceanOptics::read_ExposureTime(): error in GetParameters()" << endl;
	if(get_state() == Tango::ON)
	{
	int error;
#if _SEABREEZE_API > 1
	unsigned long min_integration_time_microsec=0;
#else
	char error_str[128];
	long min_integration_time_microsec;/* = seabreeze_get_min_integration_time_microsec(device_id, &error);
	seabreeze_get_error_string(error, error_str, 128);
	printf("min_integration_time_microsec is %d [%s]\n", min_integration_time_microsec, error_str);*/
	const int SEND_ENDPOINT = 0x01;
	const int READ_ENDPOINT = 0x81;
	unsigned char request[1];
	unsigned char response[32];
	min_integration_time_microsec = 0;
	size_t i = 0;
	//for(i=0; i<1; i++)
	{
		request[0] = 0xfe;
		//request[1] = 2+i;

		memset(response, 0, sizeof(response));

		seabreeze_write_usb(device_id, &error, SEND_ENDPOINT, request, sizeof(request));
		if(error)
		{
			seabreeze_get_error_string(error, error_str, 128);
			printf("Error writing usb [%s]\n", error_str);
			set_state(Tango::FAULT);
			set_status("Communication error");
			disconnect_device();
			device_id = -1;
			ostringstream str;
			str << "Error: '" << error_str << "'";
			Tango::Except::throw_exception(
				(const char *) "Communication error",
				str.str(),
				__func__, Tango::ERR);
			//continue;
		}

		seabreeze_read_usb(device_id, &error, READ_ENDPOINT, response, 17/*sizeof(response)*/);
		if(error)
		{
			seabreeze_get_error_string(error, error_str, 128);
			printf("Error reading usb [%s]\n", error_str);
			set_state(Tango::FAULT);
			set_status("Communication error");
			disconnect_device();
			ostringstream str;
			str << "Error: '" << error_str << "'";
			Tango::Except::throw_exception(
				(const char *) "Communication error",
				str.str(),
				__func__, Tango::ERR);
		}

		/*unsigned short value = (unsigned short) (response[1] | (response[2] << 8));

		printf("i=%d: hex 0x%04x dec %5hu (%s)\n", i, value, value, labels[i]);

		i++;*/
		//min_integration_time_microsec = (long) (response[1] << (i*8));
    }
	//printf("response 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n", response[0], response[1], response[2], response[3], response[4], response[5], response[6], response[7], response[8], response[9], response[10], response[11], response[12], response[13], response[14]);
	//min_integration_time_microsec = (long) (response[2] | (response[3] << 8) | (response[4] << 16) | (response[5] << 24));
	uint32_t *tmp = (uint32_t *)&response[2];
	min_integration_time_microsec = (long)(myntohl(*tmp));
#endif
	*attr_ExposureTime_read = (double)min_integration_time_microsec/1000;//TODO
	}
	attr.set_value(attr_ExposureTime_read);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_ExposureTime
}
//--------------------------------------------------------
/**
 *	Write attribute ExposureTime related method
 *	Description: Exposure time in milliseconds
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_ExposureTime(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_ExposureTime(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_ExposureTime) ENABLED START -----*/
	if(get_state() == Tango::FAULT)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			(const char *) "OceanOptics::write_ExposureTime()", Tango::ERR);
	}
	if(get_state() != Tango::ON && get_state() != Tango::ALARM)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			(const char *) "OceanOptics::write_ExposureTime()", Tango::ERR);
	}
	int error;

#if _SEABREEZE_API > 1
	unsigned long integr_time_usec = (unsigned long)(w_val*1000);
	sbapi_spectrometer_set_integration_time_micros (device_id, spectrometer_id, &error, integr_time_usec);
#else
	long integr_time_usec = (long)(w_val*1000);
    seabreeze_set_integration_time_microsec(device_id, &error, integr_time_usec);
#endif
	if(error)
	{
		char error_str[128];
#if _SEABREEZE_API > 1
#else
		seabreeze_get_error_string(error, error_str, 128);
#endif
		printf("Error is is %d [%s]\n", error, error_str);
		TangoSys_MemStream out_stream;
		out_stream << "Error writing Exposure Time=" << w_val << ends;
		Tango::Except::throw_exception(
			(const char *) "Value not Allowed",
			out_stream.str(),
			(const char *) "OceanOptics::write_ExposureTime()", Tango::ERR);
	}

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_ExposureTime
}
//--------------------------------------------------------
/**
 *	Read attribute IncrSensitivity related method
 *	Description: false: use standard sensitivity\ntrue: use increased sensitivity
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_IncrSensitivity(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_IncrSensitivity(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_IncrSensitivity) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_IncrSensitivity(Tango::Attribute &attr) entering... "<< endl;
	//if(!m_usbCCD.GetParameters())
	//	INFO_STREAM << "OceanOptics::read_IncrSensitivity(): error in GetParameters()" << endl;
	//DEBUG_STREAM << "OceanOptics::read_IncrSensitivity() read sensitivity=" << m_usbCCD.m_dwSensitivity << endl;
	*attr_IncrSensitivity_read = 0; //TODO: m_usbCCD.m_dwSensitivity != 0;
	attr.set_value(attr_IncrSensitivity_read);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_IncrSensitivity
}
//--------------------------------------------------------
/**
 *	Write attribute IncrSensitivity related method
 *	Description: false: use standard sensitivity\ntrue: use increased sensitivity
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_IncrSensitivity(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_IncrSensitivity(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_IncrSensitivity) ENABLED START -----*/
	if(get_state() == Tango::FAULT)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			(const char *) "OceanOptics::write_IncrSensitivity()", Tango::ERR);
	}
	if(get_state() != Tango::ON)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			(const char *) "OceanOptics::write_IncrSensitivity()", Tango::ERR);
	}
	/*if(!(properties & DP_SENSIT))
	{
		TangoSys_MemStream out_stream;
		out_stream << "Changing Increased Sensitivity not supported on this model" << ends;
		Tango::Except::throw_exception(
			(const char *) "Not supported",
			out_stream.str(),
			(const char *) "OceanOptics::write_IncrSensitivity()", Tango::ERR);
	}	if(!m_usbCCD.SetReadOuts((int)w_val))
	{
		TangoSys_MemStream out_stream;
		out_stream << "Error writing Increased Sensitivity=" << w_val << ends;
		Tango::Except::throw_exception(
			(const char *) "Value not Allowed",
			out_stream.str(),
			(const char *) "OceanOptics::write_IncrSensitivity()", Tango::ERR);
	}*/

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_IncrSensitivity
}
//--------------------------------------------------------
/**
 *	Read attribute SyncMode related method
 *	Description: 0: The camera operates without synchronization.
 *               Spectra are read out periodically in accordance with NumReadouts attribute.
 *               
 *               1: Readout is performed periodically in accordance with NumReadouts attribute.
 *               At the beginning of the first readout a positive sync pulse is formed.
 *               
 *               2: A sync pulse is formed at beginning of each readout.
 *               
 *               3: One sync pulse of negative polarity is generated for the entire readout time.
 *               
 *               4: The camera will wait for one sync pulse.
 *               The first readout will start at the rising edge of an external sync pulse.
 *               All other readouts will be performed as soon as possible.
 *               No blanks within limit
 *               
 *               5: Each readout starts at the rising edge of an external sync pulse.
 *               The camera will wait fo as many sync pulses as much readouts are set.
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_SyncMode(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_SyncMode(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_SyncMode) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_SyncMode(Tango::Attribute &attr) entering... "<< endl;
	//if(!m_usbCCD.GetParameters())
	//	INFO_STREAM << "OceanOptics::read_SyncMode(): error in GetParameters()" << endl;
	/*if(get_state() != Tango::ON)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			(const char *) __func__, Tango::ERR);
	}*/
	/*switch(m_usbCCD.m_dwSynchr)
	{
		case SYNCHR_NONE:
			*attr_SyncMode_read = TG_SYNCHR_NONE;
			break;
		case SYNCHR_CONTR:
			*attr_SyncMode_read = TG_SYNCHR_CONTR;
			break;
		case SYNCHR_CONTR_FRS:
			*attr_SyncMode_read = TG_SYNCHR_CONTR_FRS;
			break;
		case SYNCHR_CONTR_NEG:
			*attr_SyncMode_read = TG_SYNCHR_CONTR_NEG;
			break;
		case SYNCHR_EXT:
			*attr_SyncMode_read = TG_SYNCHR_EXT;
			break;
		case SYNCHR_EXT_FRS:
			*attr_SyncMode_read = TG_SYNCHR_EXT_FRS;
			break;
		default:
			*attr_SyncMode_read = -1;
			break;
	}*/
	int error;
#if 0	//TODO: not working
#if _SEABREEZE_API > 1

#else
	char error_str[128];
	uint8_t syncmode;
	const int SEND_ENDPOINT = 0x01;
	const int READ_ENDPOINT = 0x81;
	unsigned char request[1];
	unsigned char response[2];

	unsigned i = 0;
	for(i=0; i<1; i++)
	{
		//request[0] = 0xfe;
		request[0] = 0x0b;
		//request[1] = 2+i;

		memset(response, 0, sizeof(response));

		seabreeze_write_usb(device_id, &error, SEND_ENDPOINT, request, sizeof(request));
		if(error)
		{
			seabreeze_get_error_string(error, error_str, 128);
			printf("Error writing usb [%s]\n", error_str);
			continue;
		}

		seabreeze_read_usb(device_id, &error, READ_ENDPOINT, response, 2/*sizeof(response)*/);
		if(error)
		{
			seabreeze_get_error_string(error, error_str, 128);
			printf("Error reading usb [%s]\n", error_str);
			continue;
		}

		/*unsigned short value = (unsigned short) (response[1] | (response[2] << 8));

		printf("i=%d: hex 0x%04x dec %5hu (%s)\n", i, value, value, labels[i]);

		i++;*/
		//min_integration_time_microsec = (long) (response[1] << (i*8));
    }
	printf("response 0x%x 0x%x\n", response[0], response[1]);
	//min_integration_time_microsec = (long) (response[2] | (response[3] << 8) | (response[4] << 16) | (response[5] << 24));
	syncmode = response[0];
	*attr_SyncMode_read = syncmode;
#endif
#endif
	attr.set_value(attr_SyncMode_read);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_SyncMode
}
//--------------------------------------------------------
/**
 *	Write attribute SyncMode related method
 *	Description: 0: The camera operates without synchronization.
 *               Spectra are read out periodically in accordance with NumReadouts attribute.
 *               
 *               1: Readout is performed periodically in accordance with NumReadouts attribute.
 *               At the beginning of the first readout a positive sync pulse is formed.
 *               
 *               2: A sync pulse is formed at beginning of each readout.
 *               
 *               3: One sync pulse of negative polarity is generated for the entire readout time.
 *               
 *               4: The camera will wait for one sync pulse.
 *               The first readout will start at the rising edge of an external sync pulse.
 *               All other readouts will be performed as soon as possible.
 *               No blanks within limit
 *               
 *               5: Each readout starts at the rising edge of an external sync pulse.
 *               The camera will wait fo as many sync pulses as much readouts are set.
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_SyncMode(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_SyncMode(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_SyncMode) ENABLED START -----*/
	int error;
	if(get_state() != Tango::ON && get_state() != Tango::ALARM)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		DEBUG_STREAM << "OceanOptics::write_SyncMode(Tango::WAttribute &attr): " << out_stream.str() << endl;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			(const char *) "OceanOptics::write_SyncMode()", Tango::ERR);
	}
	Tango::DevLong	tmp_SyncMode;
	/*switch(w_val)
	{
		case TG_SYNCHR_NONE:
			tmp_SyncMode = SYNCHR_NONE;
			break;
		case TG_SYNCHR_CONTR:
			if(!(properties & DP_SYNCHR_CONTR))
			{
				TangoSys_MemStream out_stream;
				out_stream << "Sync Mode not supported on this model" << ends;
				Tango::Except::throw_exception(
					(const char *) "Not supported",
					out_stream.str(),
					(const char *) "OceanOptics::write_SyncMode()", Tango::ERR);
			}
			tmp_SyncMode = SYNCHR_CONTR;
			break;
		case TG_SYNCHR_CONTR_FRS:
			if(!(properties & DP_SYNCHR_CONTR_FRS))
			{
				TangoSys_MemStream out_stream;
				out_stream << "Sync Mode not supported on this model" << ends;
				Tango::Except::throw_exception(
					(const char *) "Not supported",
					out_stream.str(),
					(const char *) "OceanOptics::write_SyncMode()", Tango::ERR);
			}
			tmp_SyncMode = SYNCHR_CONTR_FRS;
			break;
		case TG_SYNCHR_CONTR_NEG:
			if(!(properties & DP_SYNCHR_CONTR_NEG))
			{
				TangoSys_MemStream out_stream;
				out_stream << "Sync Mode not supported on this model" << ends;
				Tango::Except::throw_exception(
					(const char *) "Not supported",
					out_stream.str(),
					(const char *) "OceanOptics::write_SyncMode()", Tango::ERR);
			}
			tmp_SyncMode = SYNCHR_CONTR_NEG;
			break;
		case TG_SYNCHR_EXT:
			if(!(properties & DP_SYNCHR_EXT))
			{
				TangoSys_MemStream out_stream;
				out_stream << "Sync Mode not supported on this model" << ends;
				Tango::Except::throw_exception(
					(const char *) "Not supported",
					out_stream.str(),
					(const char *) "OceanOptics::write_SyncMode()", Tango::ERR);
			}
			tmp_SyncMode = SYNCHR_EXT;
			break;
		case TG_SYNCHR_EXT_FRS:
			if(!(properties & DP_SYNCHR_EXT_FRS))
			{
				TangoSys_MemStream out_stream;
				out_stream << "Sync Mode not supported on this model" << ends;
				Tango::Except::throw_exception(
					(const char *) "Not supported",
					out_stream.str(),
					(const char *) "OceanOptics::write_SyncMode()", Tango::ERR);
			}
			tmp_SyncMode = SYNCHR_EXT_FRS;
			break;
		default:
		{
			TangoSys_MemStream out_stream;
			out_stream << "Allowed values: 0=SYNCHR_NONE, 1=SYNCHR_CONTR, 2=SYNCHR_CONTR_FRS, 3=SYNCHR_CONTR_NEG, 4=SYNCHR_EXT, 5=SYNCHR_EXT_FRS" << ends;
			Tango::Except::throw_exception(
				(const char *) "Value Not Allowed",
				out_stream.str(),
				(const char *) "OceanOptics::write_SyncMode()", Tango::ERR);
		}
	}*/
#if _SEABREEZE_API > 1
#else
	DEBUG_STREAM << "OceanOptics::write_SyncMode(Tango::WAttribute &attr): val=" << w_val << endl;
	seabreeze_set_trigger_mode(device_id, &error, w_val);
	DEBUG_STREAM << "OceanOptics::write_SyncMode(Tango::WAttribute &attr): written , error=" << error << endl;
#endif
	if(error)
	{
		char error_str[128];
	    seabreeze_get_error_string(error, error_str, 128);
	    printf("Result is [%s]\n", error_str);
		TangoSys_MemStream out_stream;
		out_stream << "Error writing Synchronization Mode=" << w_val <<" err="<<error_str;
		DEBUG_STREAM << "OceanOptics::write_SyncMode(Tango::WAttribute &attr): " << out_stream.str() << endl;
		Tango::Except::throw_exception(
			(const char *) "Value not Allowed",
			out_stream.str(),
			(const char *) "OceanOptics::write_SyncMode()", Tango::ERR);
	}
	else
	{
		*attr_SyncMode_read = w_val;
	}

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_SyncMode
}
//--------------------------------------------------------
/**
 *	Read attribute NoiseSubtraction related method
 *	Description: Enable/Disable subtraction of background noise profile \nacquired with the AcquireNoise command
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_NoiseSubtraction(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_NoiseSubtraction(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_NoiseSubtraction) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_NoiseSubtraction(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(attr_NoiseSubtraction_read);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_NoiseSubtraction
}
//--------------------------------------------------------
/**
 *	Write attribute NoiseSubtraction related method
 *	Description: Enable/Disable subtraction of background noise profile \nacquired with the AcquireNoise command
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_NoiseSubtraction(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_NoiseSubtraction(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_NoiseSubtraction) ENABLED START -----*/
	*attr_NoiseSubtraction_read = w_val;

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_NoiseSubtraction
}
//--------------------------------------------------------
/**
 *	Read attribute GaussFitEnable related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_GaussFitEnable(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_GaussFitEnable(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_GaussFitEnable) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_GaussFitEnable(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(attr_GaussFitEnable_read);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_GaussFitEnable
}
//--------------------------------------------------------
/**
 *	Write attribute GaussFitEnable related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_GaussFitEnable(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_GaussFitEnable(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_GaussFitEnable) ENABLED START -----*/
	*attr_GaussFitEnable_read = w_val;

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_GaussFitEnable
}
//--------------------------------------------------------
/**
 *	Read attribute FWHMEnable related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_FWHMEnable(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_FWHMEnable(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_FWHMEnable) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_FWHMEnable(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(attr_FWHMEnable_read);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_FWHMEnable
}
//--------------------------------------------------------
/**
 *	Write attribute FWHMEnable related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_FWHMEnable(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_FWHMEnable(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_FWHMEnable) ENABLED START -----*/
	*attr_FWHMEnable_read = w_val;

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_FWHMEnable
}
//--------------------------------------------------------
/**
 *	Read attribute GaussFitCenter related method
 *	Description: wavelength of peak of Gaussian Fit
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_GaussFitCenter(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_GaussFitCenter(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_GaussFitCenter) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_GaussFitCenter(Tango::Attribute &attr) entering... "<< endl;
	if(GaussFit_valid)
	{
		attr.set_value(attr_GaussFitCenter_read);
	}
	else
	{
		attr.set_quality(Tango::ATTR_INVALID);
	}

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_GaussFitCenter
}
//--------------------------------------------------------
/**
 *	Read attribute GaussFitWidth related method
 *	Description: width of Gaussian Fit
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_GaussFitWidth(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_GaussFitWidth(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_GaussFitWidth) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_GaussFitWidth(Tango::Attribute &attr) entering... "<< endl;
	if(GaussFit_valid)
	{
		attr.set_value(attr_GaussFitWidth_read);
	}
	else
	{
		attr.set_quality(Tango::ATTR_INVALID);
	}

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_GaussFitWidth
}
//--------------------------------------------------------
/**
 *	Read attribute FWHMCenter related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_FWHMCenter(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_FWHMCenter(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_FWHMCenter) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_FWHMCenter(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(attr_FWHMCenter_read);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_FWHMCenter
}
//--------------------------------------------------------
/**
 *	Read attribute FWHMWidth related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_FWHMWidth(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_FWHMWidth(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_FWHMWidth) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_FWHMWidth(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(attr_FWHMWidth_read);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_FWHMWidth
}
//--------------------------------------------------------
/**
 *	Read attribute Center related method
 *	Description: wavelength of peak of FWHM or of Gaussian Fit
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_Center(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_Center(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_Center) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_Center(Tango::Attribute &attr) entering... "<< endl;
	if(*attr_FWHMEnable_read || (*attr_GaussFitEnable_read && GaussFit_valid))
		attr.set_value(attr_Center_read);
	else
		attr.set_quality(Tango::ATTR_INVALID);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_Center
}
//--------------------------------------------------------
/**
 *	Read attribute Width related method
 *	Description: width of FWHM or of Gaussian Fit
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_Width(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_Width(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_Width) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_Width(Tango::Attribute &attr) entering... "<< endl;
	if(*attr_FWHMEnable_read || (*attr_GaussFitEnable_read && GaussFit_valid))
		attr.set_value(attr_Width_read);
	else
		attr.set_quality(Tango::ATTR_INVALID);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_Width
}
//--------------------------------------------------------
/**
 *	Read attribute FitType related method
 *	Description: 1=Gaussian 2=Gaussian Asymm 3=Confiteor
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_FitType(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_FitType(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_FitType) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_FitType(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(attr_FitType_read);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_FitType
}
//--------------------------------------------------------
/**
 *	Write attribute FitType related method
 *	Description: 1=Gaussian 2=Gaussian Asymm 3=Confiteor
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_FitType(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_FitType(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_FitType) ENABLED START -----*/
	*attr_FitType_read = w_val;

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_FitType
}
//--------------------------------------------------------
/**
 *	Read attribute FWHMAvg related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_FWHMAvg(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_FWHMAvg(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_FWHMAvg) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_FitType(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(attr_FWHMAvg_read);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_FWHMAvg
}
//--------------------------------------------------------
/**
 *	Write attribute FWHMAvg related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_FWHMAvg(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_FWHMAvg(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_FWHMAvg) ENABLED START -----*/
	*attr_FWHMAvg_read = w_val;

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_FWHMAvg
}
//--------------------------------------------------------
/**
 *	Read attribute NoiseLevel related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_NoiseLevel(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_NoiseLevel(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_NoiseLevel) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_NoiseLevel(Tango::Attribute &attr) entering... "<< endl;
	*attr_NoiseLevel_read = noise_level;
	attr.set_value(attr_NoiseLevel_read);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_NoiseLevel
}
//--------------------------------------------------------
/**
 *	Write attribute NoiseLevel related method
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_NoiseLevel(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_NoiseLevel(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevLong	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_NoiseLevel) ENABLED START -----*/
	noise_level = w_val;

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_NoiseLevel
}
//--------------------------------------------------------
/**
 *	Read attribute DataReferenceFile related method
 *	Description: 
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_DataReferenceFile(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_DataReferenceFile(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_DataReferenceFile) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_DataReferenceFile(Tango::Attribute &attr) entering... "<< endl;
	attr_tmp_ptr = attr_strchar;
	attr_DataReferenceFile_read = &attr_tmp_ptr;
	attr.set_value(attr_DataReferenceFile_read);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_DataReferenceFile
}
//--------------------------------------------------------
/**
 *	Write attribute DataReferenceFile related method
 *	Description: 
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_DataReferenceFile(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_DataReferenceFile(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevString	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_DataReferenceFile) ENABLED START -----*/
	ostringstream tempstr;

	tempstr << w_val << ends;
	strcpy(attr_strchar,tempstr.str().c_str());
	if(strlen(attr_strchar) > 0)
	{
		dataReferenceFile  = tempstr.str();

		ifstream  fd;		//startup file
		fd.open(dataReferenceFile.c_str());
		if (!fd)
		{
			stringstream tmp;
			tmp << "Cannot open Data Reference file:" << strerror(errno) << endl;
			INFO_STREAM << "OceanOptics::write_DataReferenceFile(): " << tmp.str();
			Tango::Except::throw_exception(
				(const char *) "Error opening file",
				tmp.str(),
				(const char *) "OceanOptics::write_DataReferenceFile()", Tango::ERR);
   		}
		else{
			DEBUG_STREAM << "OceanOptics::write_DataReferenceFile(): file opened successfully" << endl;
			string lineread;
			int i=0;
			double val;
			while(std::getline(fd, lineread, '\n'))
			{ // Read line by line
				if(lineread.length() > 0 && i < NumPixel)
				{
					sscanf(lineread.c_str(),"%lf",&val);
					pReferenceData[i++] = val;
				}
			}
			fd.close();
		}
	}

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_DataReferenceFile
}
//--------------------------------------------------------
/**
 *	Read attribute AutoExposure related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_AutoExposure(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_AutoExposure(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_AutoExposure) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_AutoExposure(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(attr_AutoExposure_read);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_AutoExposure
}
//--------------------------------------------------------
/**
 *	Write attribute AutoExposure related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_AutoExposure(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_AutoExposure(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_AutoExposure) ENABLED START -----*/
	*attr_AutoExposure_read = w_val;
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_AutoExposure
}
//--------------------------------------------------------
/**
 *	Read attribute LightSourceIntensity related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_LightSourceIntensity(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_LightSourceIntensity(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_LightSourceIntensity) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_LightSourceIntensity_read);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_LightSourceIntensity
}
//--------------------------------------------------------
/**
 *	Write attribute LightSourceIntensity related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_LightSourceIntensity(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_LightSourceIntensity(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_LightSourceIntensity) ENABLED START -----*/

	if(get_state() == Tango::FAULT)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			__func__, Tango::ERR);
	}
	if(get_state() != Tango::ON && get_state() != Tango::ALARM)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			__func__, Tango::ERR);
	}
	int error;

#if _SEABREEZE_API > 1
	unsigned long integr_time_usec = (unsigned long)(w_val*1000);
	sbapi_spectrometer_set_integration_time_micros (device_id, spectrometer_id, &error, integr_time_usec);
#else
	int count_light_id = seabreeze_get_light_source_count (device_id, &error);
	if(count_light_id != 1)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Count of light sources different from 1 but is " << count_light_id;
		Tango::Except::throw_exception(
			(const char *) "Bad light sources count",
			out_stream.str(),
			__func__, Tango::ERR);
	}
	seabreeze_set_light_source_intensity(device_id, &error, 0/*light_id*/, w_val);
#endif
	if(error)
	{
		char error_str[128];
#if _SEABREEZE_API > 1
#else
		seabreeze_get_error_string(error, error_str, 128);
#endif
		printf("Error is is %d [%s]\n", error, error_str);
		TangoSys_MemStream out_stream;
		out_stream << "Error writing seabreeze_set_light_source_intensity=" << w_val << " err=" << error_str;
		Tango::Except::throw_exception(
			(const char *) "Error",
			out_stream.str(),
			__func__, Tango::ERR);
	}
	*attr_LightSourceIntensity_read = w_val;


	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_LightSourceIntensity
}
//--------------------------------------------------------
/**
 *	Read attribute RtMode related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_RtMode(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_RtMode(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_RtMode) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_RtMode_read);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_RtMode
}
//--------------------------------------------------------
/**
 *	Write attribute RtMode related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_RtMode(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_RtMode(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_RtMode) ENABLED START -----*/
	*attr_RtMode_read = w_val;

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_RtMode
}
//--------------------------------------------------------
/**
 *	Read attribute LightSourceEnable related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_LightSourceEnable(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_LightSourceEnable(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_LightSourceEnable) ENABLED START -----*/
	if(get_state() == Tango::ON)
	{
	int error;
#if _SEABREEZE_API > 1
	unsigned long min_integration_time_microsec=0;
#else
	char error_str[128];
	long min_integration_time_microsec;/* = seabreeze_get_min_integration_time_microsec(device_id, &error);
	seabreeze_get_error_string(error, error_str, 128);
	printf("min_integration_time_microsec is %d [%s]\n", min_integration_time_microsec, error_str);*/
	const int SEND_ENDPOINT = 0x01;
	const int READ_ENDPOINT = 0x81;
	unsigned char request[1];
	unsigned char response[32];
	min_integration_time_microsec = 0;
	size_t i = 0;
	//for(i=0; i<1; i++)
	{
		request[0] = 0xfe;
		//request[1] = 2+i;

		memset(response, 0, sizeof(response));

		seabreeze_write_usb(device_id, &error, SEND_ENDPOINT, request, sizeof(request));
		if(error)
		{
			seabreeze_get_error_string(error, error_str, 128);
			printf("Error writing usb [%s]\n", error_str);
			set_state(Tango::FAULT);
			set_status("Communication error");
			disconnect_device();
			device_id = -1;
			ostringstream str;
			str << "Error: '" << error_str << "'";
			Tango::Except::throw_exception(
				(const char *) "Communication error",
				str.str(),
				__func__, Tango::ERR);
		}

		seabreeze_read_usb(device_id, &error, READ_ENDPOINT, response, 17/*sizeof(response)*/);
		if(error)
		{
			seabreeze_get_error_string(error, error_str, 128);
			printf("Error reading usb [%s]\n", error_str);
			set_state(Tango::FAULT);
			set_status("Communication error");
			disconnect_device();
			device_id = -1;
			ostringstream str;
			str << "Error: '" << error_str << "'";
			Tango::Except::throw_exception(
				(const char *) "Communication error",
				str.str(),
				__func__, Tango::ERR);
		}

		/*unsigned short value = (unsigned short) (response[1] | (response[2] << 8));

		printf("i=%d: hex 0x%04x dec %5hu (%s)\n", i, value, value, labels[i]);

		i++;*/
		//min_integration_time_microsec = (long) (response[1] << (i*8));
    }
	//printf("response 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n", response[0], response[1], response[2], response[3], response[4], response[5], response[6], response[7], response[8], response[9], response[10], response[11], response[12], response[13], response[14]);
	//min_integration_time_microsec = (long) (response[2] | (response[3] << 8) | (response[4] << 16) | (response[5] << 24));
	uint8_t *tmp = (uint8_t *)&response[8];
#endif
	*attr_LightSourceEnable_read = (tmp != 0);//TODO
	}
	//	Set the attribute value
	attr.set_value(attr_LightSourceEnable_read);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_LightSourceEnable
}
//--------------------------------------------------------
/**
 *	Write attribute LightSourceEnable related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_LightSourceEnable(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_LightSourceEnable(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_LightSourceEnable) ENABLED START -----*/
	if(get_state() == Tango::FAULT)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			__func__, Tango::ERR);
	}
	if(get_state() != Tango::ON && get_state() != Tango::ALARM)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			__func__, Tango::ERR);
	}
	int error;
#if _SEABREEZE_API > 1
	//unsigned long integr_time_usec = (unsigned long)(w_val*1000);
	//sbapi_spectrometer_set_integration_time_micros (device_id, spectrometer_id, &error, integr_time_usec);
#else
	int count_light_id = seabreeze_get_light_source_count (device_id, &error);
	if(count_light_id != 1)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Count of light sources different from 1 but is " << count_light_id;
		Tango::Except::throw_exception(
			(const char *) "Bad light sources count",
			out_stream.str(),
			__func__, Tango::ERR);
	}
	unsigned char enable = w_val  ? 1 : 0;
	seabreeze_set_light_source_enable(device_id, &error, 0/*light_id*/, enable);
#endif
	if(error)
	{
		char error_str[128];
#if _SEABREEZE_API > 1
#else
		seabreeze_get_error_string(error, error_str, 128);
#endif
		printf("Error is is %d [%s]\n", error, error_str);
		TangoSys_MemStream out_stream;
		out_stream << "Error writing seabreeze_set_light_source_enable=" << enable << " err=" << error_str;
		Tango::Except::throw_exception(
			(const char *) "Error",
			out_stream.str(),
			__func__, Tango::ERR);
	}


	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_LightSourceEnable
}
//--------------------------------------------------------
/**
 *	Read attribute IntensityCalibrationEnable related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_IntensityCalibrationEnable(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_IntensityCalibrationEnable(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_IntensityCalibrationEnable) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_IntensityCalibrationEnable_read);
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_IntensityCalibrationEnable
}
//--------------------------------------------------------
/**
 *	Write attribute IntensityCalibrationEnable related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_IntensityCalibrationEnable(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_IntensityCalibrationEnable(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_IntensityCalibrationEnable) ENABLED START -----*/
	*attr_IntensityCalibrationEnable_read = w_val;
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_IntensityCalibrationEnable
}
//--------------------------------------------------------
/**
 *	Read attribute Area related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_Area(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_Area(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_Area) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_Area_read);
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_Area
}
//--------------------------------------------------------
/**
 *	Read attribute AreaRegion related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_AreaRegion(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_AreaRegion(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_AreaRegion) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_AreaRegion_read);
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_AreaRegion
}
//--------------------------------------------------------
/**
 *	Write attribute AreaRegion related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_AreaRegion(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_AreaRegion(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_AreaRegion) ENABLED START -----*/
	*attr_AreaRegion_read = w_val;
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_AreaRegion
}
//--------------------------------------------------------
/**
 *	Read attribute AreaEnable related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_AreaEnable(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_AreaEnable(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_AreaEnable) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_GaussFitEnable(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(attr_AreaEnable_read);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_AreaEnable
}
//--------------------------------------------------------
/**
 *	Write attribute AreaEnable related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_AreaEnable(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_AreaEnable(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_AreaEnable) ENABLED START -----*/
	*attr_AreaEnable_read = w_val;

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_AreaEnable
}
//--------------------------------------------------------
/**
 *	Read attribute WidhtCorrectionEnable related method
 *	Description: enable widht correction
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::read_WidhtCorrectionEnable(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_WidhtCorrectionEnable(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_WidhtCorrectionEnable) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_WidhtCorrectionEnable_read);
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_WidhtCorrectionEnable
}
//--------------------------------------------------------
/**
 *	Write attribute WidhtCorrectionEnable related method
 *	Description: enable widht correction
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void OceanOptics::write_WidhtCorrectionEnable(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_WidhtCorrectionEnable(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_WidhtCorrectionEnable) ENABLED START -----*/
	*attr_WidhtCorrectionEnable_read = w_val;
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_WidhtCorrectionEnable
}
//--------------------------------------------------------
/**
 *	Read attribute NumPixels related method
 *	Description: long[0]: number of pixels horizontally (number of CCD-array columns)\nlong[1]: number of pixels vertically (number of CCD-array lines)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Spectrum max = 2
 */
//--------------------------------------------------------
void OceanOptics::read_NumPixels(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_NumPixels(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_NumPixels) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_NumPixels(Tango::Attribute &attr) entering... "<< endl;
	//if(!m_usbCCD.GetParameters())
	//	INFO_STREAM << "OceanOptics::read_NumPixels(): error in GetParameters()" << endl;
	//attr_NumPixels_read[0] = m_usbCCD.m_dwDigitCapacity;		//TODO: returns 12 bit
	/*attr_NumPixels_read[0] = m_usbCCD.m_nNumPixelsH;
	attr_NumPixels_read[1] = m_usbCCD.m_nNumPixelsV;*/
	attr.set_value(attr_NumPixels_read,2);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_NumPixels
}
//--------------------------------------------------------
/**
 *	Read attribute Data related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 5000
 */
//--------------------------------------------------------
void OceanOptics::read_Data(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_Data(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_Data) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_Data(Tango::Attribute &attr) entering... "<< endl;
#ifdef _NON_AVERAGE
	if(pLiveData)
	{
		attr_Data_read = (Tango::DevLong *)&(pLiveData[0 + attr_RegionOfInterest_read[0]]);
		attr.set_value(attr_Data_read,attr_RegionOfInterest_read[1] - attr_RegionOfInterest_read[0]);
	}
#else
	if(pLiveData_avg)
	{
		//DEBUG_STREAM << "OceanOptics::read_Data() ROIStartPix"<<RegionOfInterestStartPix<<" ROIEndPix="<<RegionOfInterestEndPix<<" size="<<RegionOfInterestEndPix - RegionOfInterestStartPix + 1<< endl;
		//attr_Data_read = (Tango::DevLong *)&(pLiveData_avg[0 + WaveLenght2Pix(attr_RegionOfInterest_read[0])]);
		attr_Data_read = (Tango::DevDouble *)&(pLiveData_avg[0 + RegionOfInterestStartPix]);
		//attr.set_value(attr_Data_read,WaveLenght2Pix(attr_RegionOfInterest_read[1]) - WaveLenght2Pix(attr_RegionOfInterest_read[0]));
		attr.set_value(attr_Data_read,RegionOfInterestEndPix - RegionOfInterestStartPix + 1);
	}
#endif
	else
		attr.set_quality(Tango::ATTR_INVALID);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_Data
}
//--------------------------------------------------------
/**
 *	Read attribute RegionOfInterest related method
 *	Description: Region of interest returned by the Data attribute\ndouble[0]=start of the region\ndouble[1]=end of the region\n\nValues are in nm according with the conversion factor in x_values\nData attribute property
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 2
 */
//--------------------------------------------------------
void OceanOptics::read_RegionOfInterest(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_RegionOfInterest(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_RegionOfInterest) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_RegionOfInterest(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(attr_RegionOfInterest_read,2);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_RegionOfInterest
}
//--------------------------------------------------------
/**
 *	Write attribute RegionOfInterest related method
 *	Description: Region of interest returned by the Data attribute\ndouble[0]=start of the region\ndouble[1]=end of the region\n\nValues are in nm according with the conversion factor in x_values\nData attribute property
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 2
 */
//--------------------------------------------------------
void OceanOptics::write_RegionOfInterest(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_RegionOfInterest(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve number of write values
	int	w_length = attr.get_write_value_length();

	//	Retrieve pointer on write values (Do not delete !)
	const Tango::DevDouble	*w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_RegionOfInterest) ENABLED START -----*/
	if(get_state()==Tango::FAULT)
	{
		Tango::Except::throw_exception(
			(const char *) "CCD not communicating",
			(const char *) "CCD not communicating",
			(const char *) "OceanOptics::write_RegionOfInterest()", Tango::ERR);
	}
	if(w_length != 2)
	{
		Tango::Except::throw_exception(
			(const char *) "Wrong parameters",
			(const char *) "Expected 2 parameters",
			(const char *) "OceanOptics::write_RegionOfInterest()", Tango::ERR);
	}
	Tango::DevDouble	w_val_[2];
	w_val_[0] = floor(w_val[0]*1000.0)*0.001;
	w_val_[1] = floor(w_val[1]*1000.0)*0.001;
	if(w_val_[0] < x_values[0] || w_val_[0] > x_values[NumPixel-1] ||
			w_val_[1] < x_values[0] || w_val_[1] > x_values[NumPixel-1] ||
			w_val_[0] >= w_val_[1])
	{
		TangoSys_MemStream out_stream;
		out_stream << "Region ["<<w_val_[0]<<","<<w_val_[1]<<"] not allowed, it has to be in ["<<x_values[0]<<"," << x_values[NumPixel-1] << "]" << ends;
		Tango::Except::throw_exception(
			(const char *) "Value not Allowed",
			out_stream.str(),
			(const char *) "OceanOptics::write_RegionOfInterest()", Tango::ERR);
	}
	RegionOfInterestStartPix = WaveLenght2Pix(w_val_[0]);
	DEBUG_STREAM << __func__<<": StartROI="<<w_val_[0]<<" -> start pix="<<RegionOfInterestStartPix;
	if(x_values[RegionOfInterestStartPix] < w_val_[0])
	{
		RegionOfInterestStartPix++;
	}
	DEBUG_STREAM << __func__<<": ...-> adjusted start pix="<<RegionOfInterestStartPix;
	attr_RegionOfInterest_read[0] = x_values[RegionOfInterestStartPix];
	RegionOfInterestEndPix = WaveLenght2Pix(w_val_[1]);
	DEBUG_STREAM << __func__<<": EndROI="<<w_val_[1]<<" -> end pix="<<RegionOfInterestEndPix;
	if(x_values[RegionOfInterestEndPix] > w_val_[1])
	{
		RegionOfInterestEndPix--;
	}
	DEBUG_STREAM << __func__<<": ...-> adjusted end pix="<<RegionOfInterestEndPix;
	attr_RegionOfInterest_read[1] = x_values[RegionOfInterestEndPix];
	tbuf_array_w.reset_rt_buffer();
	tbuf_array_l.reset_rt_buffer();

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_RegionOfInterest
}
//--------------------------------------------------------
/**
 *	Read attribute AnalysisRegion related method
 *	Description: Region used for gaussian and fwhm analysis\ndouble[0]=start of the region\ndouble[1]=end of the region\n\nValues are in nm according with the conversion factor in x_values\nData attribute property
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 2
 */
//--------------------------------------------------------
void OceanOptics::read_AnalysisRegion(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_AnalysisRegion(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_AnalysisRegion) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_AnalysisRegion(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(attr_AnalysisRegion_read,2);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_AnalysisRegion
}
//--------------------------------------------------------
/**
 *	Write attribute AnalysisRegion related method
 *	Description: Region used for gaussian and fwhm analysis\ndouble[0]=start of the region\ndouble[1]=end of the region\n\nValues are in nm according with the conversion factor in x_values\nData attribute property
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 2
 */
//--------------------------------------------------------
void OceanOptics::write_AnalysisRegion(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_AnalysisRegion(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve number of write values
	int	w_length = attr.get_write_value_length();

	//	Retrieve pointer on write values (Do not delete !)
	const Tango::DevDouble	*w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_AnalysisRegion) ENABLED START -----*/
	if(get_state()==Tango::FAULT)
	{
		Tango::Except::throw_exception(
			(const char *) "CCD not communicating",
			(const char *) "CCD not communicating",
			(const char *) "OceanOptics::write_AnalysisRegion()", Tango::ERR);
	}	if(w_val[0] < x_values[0] || w_val[0] > x_values[NumPixel-1] ||
		w_val[1] < x_values[0] || w_val[1] > x_values[NumPixel-1] ||
		w_val[0] >= w_val[1])
	{
		TangoSys_MemStream out_stream;
		out_stream << "Region not allowed, it has to be in ["<<x_values[0]<<"," << x_values[NumPixel-1] << "]" << ends;
		Tango::Except::throw_exception(
			(const char *) "Value not Allowed",
			out_stream.str(),
			(const char *) "OceanOptics::write_AnalysisRegion()", Tango::ERR);
	}
	AnalysisRegionStartPix = WaveLenght2Pix(w_val[0]);
	if(x_values[AnalysisRegionStartPix] < w_val[0])
	{
		AnalysisRegionStartPix++;
	}
	attr_AnalysisRegion_read[0] = x_values[AnalysisRegionStartPix];
	AnalysisRegionEndPix = WaveLenght2Pix(w_val[1]);
	if(x_values[AnalysisRegionEndPix] > w_val[1])
	{
		AnalysisRegionEndPix--;
	}
	attr_AnalysisRegion_read[1] = x_values[AnalysisRegionEndPix];

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_AnalysisRegion
}
//--------------------------------------------------------
/**
 *	Read attribute GaussFit related method
 *	Description: double[0]= wavelength of peak of Gaussian Fit\ndouble[1]= width of Gaussian Fit
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4
 */
//--------------------------------------------------------
void OceanOptics::read_GaussFit(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_GaussFit(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_GaussFit) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_GaussFit(Tango::Attribute &attr) entering... "<< endl;
	if(GaussFit_valid)
	{
		attr.set_value(attr_GaussFit_read,4);
	}
	else
	{
		attr.set_quality(Tango::ATTR_INVALID);
	}

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_GaussFit
}
//--------------------------------------------------------
/**
 *	Read attribute FWHM related method
 *	Description: double[0]= wavelength of peak of FWHM\ndouble[1]= width of FWHM
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 3
 */
//--------------------------------------------------------
void OceanOptics::read_FWHM(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_FWHM(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_FWHM) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_FWHM(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(attr_FWHM_read,3);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_FWHM
}
//--------------------------------------------------------
/**
 *	Read attribute GaussReference related method
 *	Description: Gaussian of reference params
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4
 */
//--------------------------------------------------------
void OceanOptics::read_GaussReference(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_GaussReference(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_GaussReference) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_GaussReference(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(attr_GaussReference_read,4);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_GaussReference
}
//--------------------------------------------------------
/**
 *	Write attribute GaussReference related method
 *	Description: Gaussian of reference params
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4
 */
//--------------------------------------------------------
void OceanOptics::write_GaussReference(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_GaussReference(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve number of write values
	int	w_length = attr.get_write_value_length();

	//	Retrieve pointer on write values (Do not delete !)
	const Tango::DevDouble	*w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_GaussReference) ENABLED START -----*/
	attr_GaussReference_read[0] = w_val[0];
	attr_GaussReference_read[1] = w_val[1];
	attr_GaussReference_read[2] = w_val[2];
	attr_GaussReference_read[3] = w_val[3];

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_GaussReference
}
//--------------------------------------------------------
/**
 *	Read attribute DataReference related method
 *	Description: A reference curve
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 5000
 */
//--------------------------------------------------------
void OceanOptics::read_DataReference(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_DataReference(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_DataReference) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::attr_DataReference_read(Tango::Attribute &attr) entering... "<< endl;
	if(pReferenceData)
	{
		//attr_DataReference_read = (Tango::DevLong *)&(pReferenceData[0 + WaveLenght2Pix(attr_RegionOfInterest_read[0])]);
		attr_DataReference_read = (Tango::DevDouble *)&(pReferenceData[0 + RegionOfInterestStartPix]);
		//attr.set_value(attr_DataReference_read,WaveLenght2Pix(attr_RegionOfInterest_read[1]) - WaveLenght2Pix(attr_RegionOfInterest_read[0]));
		attr.set_value(attr_DataReference_read,RegionOfInterestEndPix - RegionOfInterestStartPix + 1);
	}
	else
		attr.set_quality(Tango::ATTR_INVALID);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_DataReference
}
//--------------------------------------------------------
/**
 *	Write attribute DataReference related method
 *	Description: A reference curve
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 5000
 */
//--------------------------------------------------------
void OceanOptics::write_DataReference(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_DataReference(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve number of write values
	int	w_length = attr.get_write_value_length();

	//	Retrieve pointer on write values (Do not delete !)
	const Tango::DevDouble	*w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_DataReference) ENABLED START -----*/
	if((attr.get_write_value_length() != RegionOfInterestEndPix - RegionOfInterestStartPix + 1) &&
		(attr.get_write_value_length() != NumPixel))
	{
		ostringstream tmp;
		tmp << "Accepting attribute size = " << NumPixel << " or " << RegionOfInterestEndPix - RegionOfInterestStartPix + 1 << ends;
		Tango::Except::throw_exception(
			(const char *) "Wrong size",
			tmp.str(),
			(const char *) "OceanOptics::write_DataReference()", Tango::ERR);
	}	if(attr.get_write_value_length() == RegionOfInterestEndPix - RegionOfInterestStartPix + 1)
	{
		for(unsigned int i=0; i<RegionOfInterestStartPix; i++)
			pReferenceData[i] = 0;
		for(unsigned int i=RegionOfInterestStartPix; i< RegionOfInterestEndPix + 1; i++)
			pReferenceData[i] = w_val[i-RegionOfInterestStartPix];
		for(unsigned int i=RegionOfInterestEndPix + 1; i<NumPixel; i++)
			pReferenceData[i] = 0;
	}
	else if(attr.get_write_value_length() == NumPixel)
	{
		for(unsigned int i=0; i< NumPixel; i++)
			pReferenceData[i] = w_val[i];
	}

	ofstream  fd;		//startup file
	fd.open(dataReferenceFile.c_str());
	if (!fd)
	{
		stringstream tmp;
		tmp << "Cannot open Data Reference default file " << dataReferenceFile<< "  err:" << strerror(errno) << endl;
		INFO_STREAM << "OceanOptics::write_DataReference(): " << tmp.str();
   	}
	else
	{
		DEBUG_STREAM << "OceanOptics::write_DataReference(): file " << dataReferenceFile << " opened successfully" << endl;
		for(int i=0; i<NumPixel; i++)
		{
			fd << pReferenceData[i] << endl;
		}
		fd.close();
	}

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_DataReference
}
//--------------------------------------------------------
/**
 *	Read attribute IntensityCalibration related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 10000
 */
//--------------------------------------------------------
void OceanOptics::read_IntensityCalibration(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_IntensityCalibration(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_IntensityCalibration) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_IntensityCalibration_read, NumPixel);
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_IntensityCalibration
}
//--------------------------------------------------------
/**
 *	Write attribute IntensityCalibration related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 10000
 */
//--------------------------------------------------------
void OceanOptics::write_IntensityCalibration(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "OceanOptics::write_IntensityCalibration(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve number of write values
	int	w_length = attr.get_write_value_length();

	//	Retrieve pointer on write values (Do not delete !)
	const Tango::DevDouble	*w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(OceanOptics::write_IntensityCalibration) ENABLED START -----*/
	if(w_length != NumPixel)
	{
		ostringstream tmp;
		tmp << "Expected size is " << NumPixel << " but read " << w_length << ends;
		Tango::Except::throw_exception(
			(const char *) "Wrong size",
			tmp.str(),
			(const char *) "OceanOptics::write_IntensityCalibration()", Tango::ERR);
	}
	for(int i=0; i<w_length; i++)
	{
		attr_IntensityCalibration_read[i] = w_val[i];
	}
	save_intensity_cal_file(intensityCalibrationFile);
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::write_IntensityCalibration
}
//--------------------------------------------------------
/**
 *	Read attribute NoiseProfile related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 5000
 */
//--------------------------------------------------------
void OceanOptics::read_NoiseProfile(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_NoiseProfile(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_NoiseProfile) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_Data(Tango::Attribute &attr) entering... "<< endl;
#ifdef _NON_AVERAGE
	if(pNoiseMask)
	{
		attr_Data_read = (Tango::DevLong *)&(pNoiseMask[0 + attr_RegionOfInterest_read[0]]);
		attr.set_value(attr_NoiseMask_read,attr_RegionOfInterest_read[1] - attr_RegionOfInterest_read[0]);
	}
#else
	if(pNoiseMask_avg)
	{
		attr_NoiseProfile_read = (Tango::DevDouble *)&(pNoiseMask_avg[0 + RegionOfInterestStartPix]);
		attr.set_value(attr_NoiseProfile_read,RegionOfInterestEndPix - RegionOfInterestStartPix + 1);
	}
#endif
	else
		attr.set_quality(Tango::ATTR_INVALID);
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_NoiseProfile
}
//--------------------------------------------------------
/**
 *	Read attribute Wave related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 5000
 */
//--------------------------------------------------------
void OceanOptics::read_Wave(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_Wave(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_Wave) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_Wave(Tango::Attribute &attr) entering... "<< endl;
#ifdef _NON_AVERAGE
	if(pLiveData)
	{
		attr_Wave_read = (Tango::DevLong *)&(pLiveData[0 + attr_RegionOfInterest_read[0]]);
		attr.set_value(attr_Wave_read,attr_RegionOfInterest_read[1] - attr_RegionOfInterest_read[0]);
	}
#else
	if(pLiveData_avg)
	{
		//DEBUG_STREAM << "OceanOptics::read_Wave() ROIStartPix"<<RegionOfInterestStartPix<<" ROIEndPix="<<RegionOfInterestEndPix<<" size="<<RegionOfInterestEndPix - RegionOfInterestStartPix + 1<< endl;
		//attr_Wave_read = (Tango::DevLong *)&(pLiveData_avg[0 + WaveLenght2Pix(attr_RegionOfInterest_read[0])]);
		attr_Wave_read = (Tango::DevDouble *)&(pLiveData_avg[0 + RegionOfInterestStartPix]);
		//attr.set_value(attr_Wave_read,WaveLenght2Pix(attr_RegionOfInterest_read[1]) - WaveLenght2Pix(attr_RegionOfInterest_read[0]));
		attr.set_value(attr_Wave_read,RegionOfInterestEndPix - RegionOfInterestStartPix + 1);
	}
#endif
	else
		attr.set_quality(Tango::ATTR_INVALID);
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_Wave
}
//--------------------------------------------------------
/**
 *	Read attribute Lambda related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 5000
 */
//--------------------------------------------------------
void OceanOptics::read_Lambda(Tango::Attribute &attr)
{
	DEBUG_STREAM << "OceanOptics::read_Lambda(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(OceanOptics::read_Lambda) ENABLED START -----*/
	//DEBUG_STREAM << "OceanOptics::read_Lambda(Tango::Attribute &attr) entering... "<< endl;
#ifdef _NON_AVERAGE
	if(pLiveData)
	{
		attr_Lambda_read = (Tango::DevLong *)&(pLiveData[0 + attr_RegionOfInterest_read[0]]);
		attr.set_value(attr_Lambda_read,attr_RegionOfInterest_read[1] - attr_RegionOfInterest_read[0]);
	}
#else
	if(pLiveData_avg)
	{
		//DEBUG_STREAM << "OceanOptics::read_Data() ROIStartPix"<<RegionOfInterestStartPix<<" ROIEndPix="<<RegionOfInterestEndPix<<" size="<<RegionOfInterestEndPix - RegionOfInterestStartPix + 1<< endl;
		//attr_Lambda_read = (Tango::DevLong *)&(x_values[0 + WaveLenght2Pix(attr_RegionOfInterest_read[0])]);
		attr_Lambda_read = (Tango::DevDouble *)&(x_values[0 + RegionOfInterestStartPix]);
		//attr.set_value(attr_Lambda_read,WaveLenght2Pix(attr_RegionOfInterest_read[1]) - WaveLenght2Pix(attr_RegionOfInterest_read[0]));
		attr.set_value(attr_Lambda_read,RegionOfInterestEndPix - RegionOfInterestStartPix + 1);
	}
#endif
	else
		attr.set_quality(Tango::ATTR_INVALID);
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::read_Lambda
}

//--------------------------------------------------------
/**
 *	Method      : OceanOptics::add_dynamic_attributes()
 *	Description : Create the dynamic attributes if any
 *                for specified device.
 */
//--------------------------------------------------------
void OceanOptics::add_dynamic_attributes()
{
	/*----- PROTECTED REGION ID(OceanOptics::add_dynamic_attributes) ENABLED START -----*/

	//	Add your own code to create and add dynamic attributes if any

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::add_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Command GetVersion related method
 *	Description: Returns serial number
 *
 *	@returns serial number
 */
//--------------------------------------------------------
Tango::DevString OceanOptics::get_version()
{
	Tango::DevString argout;
	DEBUG_STREAM << "OceanOptics::GetVersion()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::get_version) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	//	Add your own code to control device here
	DEBUG_STREAM << "OceanOptics::get_version(): entering... !" << endl;
	string resp;
	bool ret = 0;//TODO: m_usbCCD.GetSerialNum(resp);
	DEBUG_STREAM << "OceanOptics::get_version(): after GetSerialNum" << endl;
	if(!ret)
	{
		INFO_STREAM << "OceanOptics::get_version() error getting serial num"  << endl;
		argout = CORBA::string_dup(" ");
	}
	else
	{
		argout = CORBA::string_dup(resp.c_str());
	}

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::get_version
	return argout;
}
//--------------------------------------------------------
/**
 *	Command Stop related method
 *	Description: Reset the CCD:
 *               stops the current readout,
 *               initiates the internal DLL variables,
 *               stops data transfer
 *
 */
//--------------------------------------------------------
void OceanOptics::stop()
{
	DEBUG_STREAM << "OceanOptics::Stop()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::stop) ENABLED START -----*/

	//	Add your own code
	if(get_state() == Tango::FAULT)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			(const char *) "OceanOptics::stop()", Tango::ERR);
	}

	//	Add your own code to control device here
	live_mode = false;
	init_live_mode = false;
	measure_mode = false;
	init_measure_mode = false;
	omni_thread::sleep(0,200000000);	//0.2 s
	//loop->mutex_sleep->lock();
	loop->pausesec = 1;
	loop->pausenano = 0;
	//loop->mutex_sleep->unlock();
	set_state(Tango::ON);
	set_status("Ready");
	/*if(!m_usbCCD.Reset())
	{
		TangoSys_MemStream out_stream;
		out_stream << "Error resetting CCD" << ends;
		Tango::Except::throw_exception(
			(const char *) "Error resetting CCD",
			out_stream.str(),
			(const char *) "OceanOptics::stop()", Tango::ERR);
	}*/

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::stop
}
//--------------------------------------------------------
/**
 *	Command StartSingle related method
 *	Description: Start single (NumReadout frames) acquisition fron the CCD
 *
 */
//--------------------------------------------------------
void OceanOptics::start_single()
{
	DEBUG_STREAM << "OceanOptics::StartSingle()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::start_single) ENABLED START -----*/

	//	Add your own code
	if(get_state() != Tango::ON)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			(const char *) "OceanOptics::start_measuring()", Tango::ERR);
	}
	set_state(Tango::RUNNING);
	set_status("Single Mode");
	live_mode = false;
	init_live_mode = false;
	acquire_noise_mode = false;
	init_acquire_noise_mode = false;
	measure_mode = true;
	init_measure_mode = true;

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::start_single
}
//--------------------------------------------------------
/**
 *	Command StartContinuous related method
 *	Description: Start acquisition in Continuous mode
 *
 */
//--------------------------------------------------------
void OceanOptics::start_continuous()
{
	DEBUG_STREAM << "OceanOptics::StartContinuous()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::start_continuous) ENABLED START -----*/

	//	Add your own code
	if(get_state() != Tango::ON)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			(const char *) "OceanOptics::start_live()", Tango::ERR);
	}
	clock_gettime(CLOCK_MONOTONIC,&last_read);
	set_state(Tango::RUNNING);
	set_status("Continuous Mode");
	loop->min_fit_time = 1000000000.0;
	loop->max_fit_time = 0;
	loop->avg_fit_time = 0;
	loop->tot_fit_time = 0;
	loop->fit_counter = 0;
	measure_mode = false;
	init_measure_mode = false;
	acquire_noise_mode = false;
	init_acquire_noise_mode = false;
	live_mode = true;
	init_live_mode = true;

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::start_continuous
}
//--------------------------------------------------------
/**
 *	Command GetLastData related method
 *	Description: returns the last argin frames saved
 *
 *	@param argin number of frames to return
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *OceanOptics::get_last_data(Tango::DevLong argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "OceanOptics::GetLastData()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::get_last_data) ENABLED START -----*/

	//	Add your own code
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------
	if(get_state() == Tango::FAULT)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			(const char *) "OceanOptics::get_last_data()", Tango::ERR);
	}
	argout  = new Tango::DevVarDoubleArray();
	DEBUG_STREAM << "OceanOptics::get_last_data(): entering... FrameBufferIndex=" << FrameBufferIndex << endl;

	//	Add your own code to control device here
	if(argin > nFramesBuffered || argin < 1)
	{
		TangoSys_MemStream out_stream;
		out_stream << "MAx number of frames=" << nFramesBuffered << ends;
		Tango::Except::throw_exception(
			(const char *) "Argin out of bounds",
			out_stream.str(),
			(const char *) "OceanOptics::get_last_data()", Tango::ERR);
	}
	//int region_size = RegionOfInterestEndPix - RegionOfInterestStartPix;
	//argout->length(argin * NumPixel);
	int j;
	DEBUG_STREAM << "OceanOptics::get_last_data() (FrameBufferIndex - argin)=" << ((int)FrameBufferIndex - argin) << endl;
	if(((int)FrameBufferIndex - argin) >= 0)
	{
		j = (int)FrameBufferIndex - argin;
	}
	else
	{
		j = nFramesBuffered - (argin - FrameBufferIndex);
	}
	int region_size = RegionOfInterestEndPix - RegionOfInterestStartPix + 1;
	argout->length(argin * region_size);
	DEBUG_STREAM << "OceanOptics::get_last_data: region size=" << region_size << " new data_counter=" << argin * region_size << endl;
	for(int i=0; i < argin; i++)
	{
		//DEBUG_STREAM << "OceanOptics::get_last_data() copying DATA! from " << j*NumPixel << " to " << i*NumPixel << endl;
		//copying only region of interest:
		//i*region_size: beginning of frame=i reduced to dimension region_size
		//j*NumPixel: beginning of original frame=j of dimension NumPixel
		//j*NumPixel + attr_RegionOfInterest_read[0]: beginning of region of interest in frame=j
		memcpy(&((*argout)[i*region_size]),&(pFrameBuffer[j*NumPixel + RegionOfInterestStartPix]), region_size*sizeof(double));
		j = (j + 1) % nFramesBuffered;
	}

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::get_last_data
	return argout;
}
//--------------------------------------------------------
/**
 *	Command AcquireNoise related method
 *	Description: Acquire argin frames of background noise
 *
 *	@param argin Number of frames to acquire as noise
 */
//--------------------------------------------------------
void OceanOptics::acquire_noise(Tango::DevLong argin)
{
	DEBUG_STREAM << "OceanOptics::AcquireNoise()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::acquire_noise) ENABLED START -----*/

	//	Add your own code
	if(get_state() != Tango::ON)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			(const char *) "OceanOptics::acquire_noise()", Tango::ERR);
	}
	if(argin <= 0)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Number of frames has to be > 0" << ends;
		Tango::Except::throw_exception(
			(const char *) "Wrong argument",
			out_stream.str(),
			(const char *) "OceanOptics::acquire_noise()", Tango::ERR);
	}
	noise_frames = argin;
	noise_frames_counter = 0;
	set_state(Tango::RUNNING);
	set_status("Acquiring noise");
	live_mode = false;
	init_live_mode = false;
	measure_mode = false;
	init_measure_mode = false;
	acquire_noise_mode = true;
	init_acquire_noise_mode = true;

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::acquire_noise
}
//--------------------------------------------------------
/**
 *	Command SaveReferenceData related method
 *	Description: Save current curve as reference curve
 *
 */
//--------------------------------------------------------
void OceanOptics::save_reference_data()
{
	DEBUG_STREAM << "OceanOptics::SaveReferenceData()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::save_reference_data) ENABLED START -----*/

	//	Add your own code
	ofstream  fd;		//startup file
	fd.open(dataReferenceFile.c_str());
	for(unsigned int i=0; i<NumPixel; i++)
	{
#ifdef _NON_AVERAGE
		if(pLiveData && pReferenceData)
		{
			pReferenceData[i] = pLiveData[i];
			if (fd)
			{
				fd << (long)pReferenceData[i] << endl;
			}
		}
#else
		if(pLiveData_avg && pReferenceData)
		{
			pReferenceData[i] = pLiveData_avg[i];
			if (fd)
			{
				fd << (double)pReferenceData[i] << endl;
			}
		}
#endif
	}

	if (!fd)
	{
		stringstream tmp;
		tmp << "Cannot save on Data Reference file " << dataReferenceFile<< "  err:" << strerror(errno) << endl;
		INFO_STREAM << "OceanOptics::save_reference_data(): " << tmp.str();
		Tango::Except::throw_exception(
			(const char *) "File opening exception",
			tmp.str(),
			(const char *) "OceanOptics::save_reference_data()", Tango::ERR);
   	}
	else
	{
		fd.close();
	}

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::save_reference_data
}
//--------------------------------------------------------
/**
 *	Command OpenShutter related method
 *	Description: 
 *
 */
//--------------------------------------------------------
void OceanOptics::open_shutter()
{
	DEBUG_STREAM << "OceanOptics::OpenShutter()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::open_shutter) ENABLED START -----*/
	
	//	Add your own code
	if(get_state() == Tango::FAULT)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			__func__, Tango::ERR);
	}
	if(get_state() != Tango::ON && get_state() != Tango::ALARM)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			__func__, Tango::ERR);
	}
	int error;

#if _SEABREEZE_API > 1
	unsigned long integr_time_usec = (unsigned long)(w_val*1000);
	sbapi_spectrometer_set_integration_time_micros (device_id, spectrometer_id, &error, integr_time_usec);
#else
	unsigned char opened = 1;
	seabreeze_set_shutter_open(device_id, &error, opened);
#endif
	if(error)
	{
		char error_str[128];
#if _SEABREEZE_API > 1
#else
		seabreeze_get_error_string(error, error_str, 128);
#endif
		printf("Error is is %d [%s]\n", error, error_str);
		TangoSys_MemStream out_stream;
		out_stream << "Error writing seabreeze_set_shutter_open=" << 1 << " err=" << error_str;
		Tango::Except::throw_exception(
			(const char *) "Error",
			out_stream.str(),
			__func__, Tango::ERR);
	}
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::open_shutter
}
//--------------------------------------------------------
/**
 *	Command CloseShutter related method
 *	Description: 
 *
 */
//--------------------------------------------------------
void OceanOptics::close_shutter()
{
	DEBUG_STREAM << "OceanOptics::CloseShutter()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::close_shutter) ENABLED START -----*/
	
	//	Add your own code
	if(get_state() == Tango::FAULT)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			__func__, Tango::ERR);
	}
	if(get_state() != Tango::ON && get_state() != Tango::ALARM)
	{
		TangoSys_MemStream out_stream;
		out_stream << "Device not ready, it is in '" << get_status() << "'" << ends;
		Tango::Except::throw_exception(
			(const char *) "Device not ready",
			out_stream.str(),
			__func__, Tango::ERR);
	}
	int error;

#if _SEABREEZE_API > 1
	unsigned long integr_time_usec = (unsigned long)(w_val*1000);
	sbapi_spectrometer_set_integration_time_micros (device_id, spectrometer_id, &error, integr_time_usec);
#else
	unsigned char opened = 0;
	seabreeze_set_shutter_open(device_id, &error, opened);
#endif
	if(error)
	{
		char error_str[128];
#if _SEABREEZE_API > 1
#else
		seabreeze_get_error_string(error, error_str, 128);
#endif
		printf("Error is is %d [%s]\n", error, error_str);
		TangoSys_MemStream out_stream;
		out_stream << "Error writing seabreeze_set_shutter_open=" << 0 << " err=" << error_str;
		Tango::Except::throw_exception(
			(const char *) "Error",
			out_stream.str(),
			__func__, Tango::ERR);
	}
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::close_shutter
}
//--------------------------------------------------------
/**
 *	Command GetData related method
 *	Description: 
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *OceanOptics::get_data(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "OceanOptics::GetData()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::get_data) ENABLED START -----*/
	
	//	Add your own code
	if(!(*attr_RtMode_read) && ((*argin)[0] == BUF_GET_BUNCH_MODE || (*argin)[0] == BUF_GET_BUNCH_UNIQUE_MODE || (*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE))
		Tango::Except::throw_exception(
				"NON RT MODE",
				"Supported only in RT mode",
				__func__, Tango::ERR);
	long buf_idx[2], num_samples;
	unsigned char unique;

	tbuffer::tcheck_get_data_param<tbuffer::_data_array<double> >(argin, &tbuf_array, buf_idx, &num_samples, &unique);
	argout  = new Tango::DevVarDoubleArray();
	num_samples *= tbuf_array.get_data_size();
	argout->length(num_samples);
	tbuffer::_data_array<double> foo;
	tbuf_array.read_data(unique, buf_idx, &(*argout)[0], &foo);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::get_data
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetGaussFitCenter related method
 *	Description: 
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *OceanOptics::get_gauss_fit_center(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "OceanOptics::GetGaussFitCenter()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::get_gauss_fit_center) ENABLED START -----*/
	
	//	Add your own code
	if(!(*attr_RtMode_read) && ((*argin)[0] == BUF_GET_BUNCH_MODE || (*argin)[0] == BUF_GET_BUNCH_UNIQUE_MODE || (*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE))
		Tango::Except::throw_exception(
				"NON RT MODE",
				"Supported only in RT mode",
				__func__, Tango::ERR);
	long buf_idx[2], num_samples;
	unsigned char unique;

	tbuffer::tcheck_get_data_param<double>(argin, &tbuf[BUF_ID_GAUSS_CENTER], buf_idx, &num_samples, &unique);
	argout  = new Tango::DevVarDoubleArray();
	argout->length(num_samples);
	tbuf[BUF_ID_GAUSS_CENTER].read_data_calib(unique, buf_idx, &(*argout)[0], 0, 1, 0);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::get_gauss_fit_center
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetGaussFitWidth related method
 *	Description: 
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *OceanOptics::get_gauss_fit_width(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "OceanOptics::GetGaussFitWidth()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::get_gauss_fit_width) ENABLED START -----*/
	
	//	Add your own code
	if(!(*attr_RtMode_read) && ((*argin)[0] == BUF_GET_BUNCH_MODE || (*argin)[0] == BUF_GET_BUNCH_UNIQUE_MODE || (*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE))
		Tango::Except::throw_exception(
				"NON RT MODE",
				"Supported only in RT mode",
				__func__, Tango::ERR);
	long buf_idx[2], num_samples;
	unsigned char unique;

	tbuffer::tcheck_get_data_param<double>(argin, &tbuf[BUF_ID_GAUSS_WIDTH], buf_idx, &num_samples, &unique);
	argout  = new Tango::DevVarDoubleArray();
	argout->length(num_samples);
	tbuf[BUF_ID_GAUSS_WIDTH].read_data_calib(unique, buf_idx, &(*argout)[0], 0, 1, 0);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::get_gauss_fit_width
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetFWHMCenter related method
 *	Description: 
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *OceanOptics::get_fwhmcenter(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "OceanOptics::GetFWHMCenter()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::get_fwhmcenter) ENABLED START -----*/
	
	//	Add your own code
	if(!(*attr_RtMode_read) && ((*argin)[0] == BUF_GET_BUNCH_MODE || (*argin)[0] == BUF_GET_BUNCH_UNIQUE_MODE || (*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE))
		Tango::Except::throw_exception(
				"NON RT MODE",
				"Supported only in RT mode",
				__func__, Tango::ERR);
	long buf_idx[2], num_samples;
	unsigned char unique;

	tbuffer::tcheck_get_data_param<double>(argin, &tbuf[BUF_ID_FWHM_CENTER], buf_idx, &num_samples, &unique);
	argout  = new Tango::DevVarDoubleArray();
	argout->length(num_samples);
	tbuf[BUF_ID_FWHM_CENTER].read_data_calib(unique, buf_idx, &(*argout)[0], 0, 1, 0);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::get_fwhmcenter
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetFWHMWidth related method
 *	Description: 
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *OceanOptics::get_fwhmwidth(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "OceanOptics::GetFWHMWidth()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::get_fwhmwidth) ENABLED START -----*/
	
	//	Add your own code
	if(!(*attr_RtMode_read) && ((*argin)[0] == BUF_GET_BUNCH_MODE || (*argin)[0] == BUF_GET_BUNCH_UNIQUE_MODE || (*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE))
		Tango::Except::throw_exception(
				"NON RT MODE",
				"Supported only in RT mode",
				__func__, Tango::ERR);
	long buf_idx[2], num_samples;
	unsigned char unique;

	tcheck_get_data_param(argin, &tbuf[BUF_ID_FWHM_WIDTH], buf_idx, &num_samples, &unique);
	argout  = new Tango::DevVarDoubleArray();
	argout->length(num_samples);
	tbuf[BUF_ID_FWHM_WIDTH].read_data_calib(unique, buf_idx, &(*argout)[0], 0, 1, 0);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::get_fwhmwidth
	return argout;
}
//--------------------------------------------------------
/**
 *	Command ExportWavelength related method
 *	Description: 
 *
 */
//--------------------------------------------------------
void OceanOptics::export_wavelength()
{
	DEBUG_STREAM << "OceanOptics::ExportWavelength()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::export_wavelength) ENABLED START -----*/
	
	//	Add your own code
	if(wavelengthAttr.size() != 2)
	{
		Tango::Except::throw_exception(
			(const char *) "Beam Status Server Not Set",
			(const char *) "Beam Status Server Not Set",
			(const char *) "SolarS100::export_wavelength()", Tango::ERR);
	}
	Tango::DeviceProxy *devbeamstatus;
	devbeamstatus = new Tango::DeviceProxy(wavelengthAttr[0]);
	try
	{
		Tango::DeviceAttribute Din(wavelengthAttr[1],*attr_Center_read);
		devbeamstatus->write_attribute(Din);
	}
	catch(Tango::DevFailed e)
	{
		delete devbeamstatus;
		stringstream tmp;
		tmp << "Failed to write attribute, err=" <<e.errors[0].desc;
		Tango::Except::re_throw_exception(e,
			(const char *) "Failed",
			tmp.str(),
			(const char *) "SolarS100::export_wavelength()", Tango::ERR);
	}
	delete devbeamstatus;
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::export_wavelength
}
//--------------------------------------------------------
/**
 *	Command GetArea related method
 *	Description: 
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *OceanOptics::get_area(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "OceanOptics::GetArea()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::get_area) ENABLED START -----*/
	
	//	Add your own code
	if(!(*attr_RtMode_read) && ((*argin)[0] == BUF_GET_BUNCH_MODE || (*argin)[0] == BUF_GET_BUNCH_UNIQUE_MODE || (*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE))
		Tango::Except::throw_exception(
				"NON RT MODE",
				"Supported only in RT mode",
				__func__, Tango::ERR);
	long buf_idx[2], num_samples;
	unsigned char unique;

	tbuffer::tcheck_get_data_param<double>(argin, &tbuf[BUF_ID_AREA], buf_idx, &num_samples, &unique);
	argout  = new Tango::DevVarDoubleArray();
	argout->length(num_samples);
	tbuf[BUF_ID_AREA].read_data_calib(unique, buf_idx, &(*argout)[0], 0, 1, 0);

	/*----- PROTECTED REGION END -----*/	//	OceanOptics::get_area
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetWave related method
 *	Description: 
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *OceanOptics::get_wave(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "OceanOptics::GetWave()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::get_wave) ENABLED START -----*/
	
	//	Add your own code
	if(!(*attr_RtMode_read) && ((*argin)[0] == BUF_GET_BUNCH_MODE || (*argin)[0] == BUF_GET_BUNCH_UNIQUE_MODE || (*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE))
		Tango::Except::throw_exception(
				"NON RT MODE",
				"Supported only in RT mode",
				__func__, Tango::ERR);
	long buf_idx[2], num_samples;
	unsigned char unique;

	tbuffer::tcheck_get_data_param<tbuffer::_data_array<double> >(argin, &tbuf_array_w, buf_idx, &num_samples, &unique);
	argout  = new Tango::DevVarDoubleArray();
	num_samples *= tbuf_array_w.get_data_size();
	argout->length(num_samples);
	tbuffer::_data_array<double> foo;
	tbuf_array_w.read_data(unique, buf_idx, &(*argout)[0], &foo);
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::get_wave
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetLambda related method
 *	Description: 
 *
 *	@param argin 
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarDoubleArray *OceanOptics::get_lambda(const Tango::DevVarLongArray *argin)
{
	Tango::DevVarDoubleArray *argout;
	DEBUG_STREAM << "OceanOptics::GetLambda()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(OceanOptics::get_lambda) ENABLED START -----*/
	
	//	Add your own code
	if(!(*attr_RtMode_read) && ((*argin)[0] == BUF_GET_BUNCH_MODE || (*argin)[0] == BUF_GET_BUNCH_UNIQUE_MODE || (*argin)[0] == BUF_GET_BUNCH_BUNCH_MODE))
		Tango::Except::throw_exception(
				"NON RT MODE",
				"Supported only in RT mode",
				__func__, Tango::ERR);
	long buf_idx[2], num_samples;
	unsigned char unique;

	tbuffer::tcheck_get_data_param<tbuffer::_data_array<double> >(argin, &tbuf_array_l, buf_idx, &num_samples, &unique);
	argout  = new Tango::DevVarDoubleArray();
	num_samples *= tbuf_array_l.get_data_size();
	argout->length(num_samples);
	tbuffer::_data_array<double> foo;
	tbuf_array_l.read_data(unique, buf_idx, &(*argout)[0], &foo);

#if 0
	size_t num_samples = RegionOfInterestEndPix - RegionOfInterestStartPix + 1;

	argout  = new Tango::DevVarDoubleArray();
	argout->length(num_samples);
	for(size_t i=0; i<num_samples; i++)
		(*argout)[i] = x_values[i+RegionOfInterestStartPix];
#endif
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::get_lambda
	return argout;
}
//--------------------------------------------------------
/**
 *	Method      : OceanOptics::add_dynamic_commands()
 *	Description : Create the dynamic commands if any
 *                for specified device.
 */
//--------------------------------------------------------
void OceanOptics::add_dynamic_commands()
{
	/*----- PROTECTED REGION ID(OceanOptics::add_dynamic_commands) ENABLED START -----*/
	
	//	Add your own code to create and add dynamic commands if any
	
	/*----- PROTECTED REGION END -----*/	//	OceanOptics::add_dynamic_commands
}

/*----- PROTECTED REGION ID(OceanOptics::namespace_ending) ENABLED START -----*/

//	Additional Methods
void OceanOptics::init_data(bool reconnection = false)
{
	/*if(!m_usbCCD.GetParameters())
		INFO_STREAM << "OceanOptics::OceanOptics(): error in GetParameters()" << endl;
	NumPixel = m_usbCCD.m_nNumPixelsH * m_usbCCD.m_nNumPixelsV;*/
	int error;
	int flag;
#if _SEABREEZE_API > 1
	NumPixel = sbapi_spectrometer_get_formatted_spectrum_length(device_id,spectrometer_id, &error);
#else
	NumPixel = seabreeze_get_formatted_spectrum_length(device_id, &error);
#endif
	attr_NumPixels_read[0] = NumPixel;
	attr_NumPixels_read[1] = 1;

	if(x_values != NULL)
	{
		delete [] x_values;
		x_values = NULL;
	}
	x_values = new Tango::DevDouble[NumPixel];
	//Tango::DbData db_data_get;
	try {
	/*	db_data_get.push_back(Tango::DbDatum("Data"));
		get_db_device()->get_attribute_property(db_data_get);
		vector<Tango::DevDouble> xval;
		for(unsigned int i=0; i<db_data_get.size(); i++)
		{
			long nb_prop;
			//string &att_name = db_data[i].name;
			db_data_get[i] >> nb_prop;
			i++;
			for(int k=0; k < nb_prop; k++)
			{
				string &prop_name = db_data_get[i].name;
				if(prop_name == "x_values")
				{
					db_data_get[i] >> xval;
					break;
				}
			}
		}
		if(xval.size() > NumPixel)
		{
			WARN_STREAM << "OceanOptics::OceanOptics(): x_values property size=" << xval.size() << " greater than frame size=" << NumPixel << endl;
		}
		else
		{
			DEBUG_STREAM << "OceanOptics::OceanOptics(): x_values property size=" << xval.size() << ", frame size=" << NumPixel << endl;
			for(unsigned int i=0; i<xval.size(); i++)
			{
				x_values[i] = (Tango::DevDouble)xval[i];
			}
			//TEMP: if missing some xval, fill continuing with linear interpulation
			if(xval.size() < NumPixel)
			{
				for(unsigned int i=xval.size(); i<NumPixel; i++)
				{
					x_values[i] = (Tango::DevDouble)(xval[xval.size()-1] + (xval[xval.size()-1] - xval[xval.size()-2])*(i-xval.size()+1));
					DEBUG_STREAM << "OceanOptics::OceanOptics(): added missing x_values: x_values["<<i<<"]=" << x_values[i] << endl;
				}
			}
		}*/
		Tango::DbDatum db_wavelen("Data");	//TODO: attribute name, property name, ...
		db_wavelen << (short)1;	//1 attribute properties in "Wavelength"
		Tango::DbDatum db_values("x_values");	//TODO: attribute name, property name, ...
		Tango::DbData db_data_set;
#if _SEABREEZE_API > 1
		/*TODO: lenght*/flag = sbapi_spectrometer_get_wavelengths(device_id, spectrometer_id, &error, x_values, NumPixel);
		if(flag != NumPixel)
		{
			INFO_STREAM << __func__ << ": sbapi_spectrometer_get_wavelengths returned length="<<flag<<" but NumPixel="<< NumPixel << endl;
		}
#else
		flag = seabreeze_get_wavelengths(device_id, &error, x_values, NumPixel);
#endif
		vector<Tango::DevDouble> xval;
		for(unsigned int i=0; i<NumPixel; i++)
		{
			x_values[i] = floor(x_values[i]*1000.0)*0.001;
			//DEBUG_STREAM << "OceanOptics::OceanOptics(): x_values["<<i<<"]=" << x_values[i] << endl;
			xval.push_back(x_values[i]);
		}
	    db_values << xval;
		db_data_set.push_back(db_wavelen);
		db_data_set.push_back(db_values);
		get_db_device()->put_attribute_property(db_data_set);
	}
	catch(Tango::DevFailed e) {
		string temp;
		temp = e.errors[0].desc;
		WARN_STREAM << "OceanOptics::OceanOptics() error setting x_values attribute property: " << temp << endl;
	}
	if(pLiveData_avg != NULL)
	{
		delete [] pLiveData_avg;
		pLiveData_avg = NULL;
	}
	pLiveData_avg = new double[NumPixel];
	memset(pLiveData_avg,0,NumPixel*sizeof(double));
	if(pNoiseMask_avg != NULL)
	{
		delete [] pNoiseMask_avg;
		pNoiseMask_avg = NULL;
	}
	pNoiseMask_avg = new double[NumPixel];
	memset(pNoiseMask_avg,0,NumPixel*sizeof(double));
	if(pReferenceData != NULL)
	{
		delete [] pReferenceData;
		pReferenceData = NULL;
	}
	pReferenceData = new double[NumPixel];
	memset(pReferenceData,0,NumPixel*sizeof(double));
	
	/*defaultEndRegion = (defaultEndRegion > (int)NumPixel) ? NumPixel :
							(defaultEndRegion < 0) ? NumPixel : defaultEndRegion;
	defaultStartRegion = (defaultStartRegion < 0) ? 0 :
								((defaultStartRegion > (int)NumPixel) ? 0 :
									((defaultStartRegion >= defaultEndRegion) ? defaultEndRegion-1 : defaultStartRegion));*/
	//attr_AnalysisRegion_read[0] = attr_RegionOfInterest_read[0] = defaultStartRegion;
	RegionOfInterestStartPix = WaveLenght2Pix(attr_RegionOfInterest_read[0]);
	DEBUG_STREAM << __func__<<": StartROI="<<defaultStartRegion<<" -> start pix="<<RegionOfInterestStartPix;
	while(x_values[RegionOfInterestStartPix] < defaultStartRegion)
	{
		RegionOfInterestStartPix++;
	}
	DEBUG_STREAM << __func__<<": ... -> adjusted start pix="<<RegionOfInterestStartPix;
	AnalysisRegionStartPix = WaveLenght2Pix(attr_AnalysisRegion_read[0]);
	//attr_AnalysisRegion_read[1] = attr_RegionOfInterest_read[1] = defaultEndRegion;
	RegionOfInterestEndPix = WaveLenght2Pix(attr_RegionOfInterest_read[1]);
	DEBUG_STREAM << __func__<<": EndROI="<<defaultEndRegion<<" -> end pix="<<RegionOfInterestEndPix;
	while(x_values[RegionOfInterestEndPix] > defaultEndRegion)
	{
		RegionOfInterestEndPix--;
	}
	DEBUG_STREAM << __func__<<": ... -> adjusted end pix="<<RegionOfInterestEndPix;
	AnalysisRegionEndPix = WaveLenght2Pix(attr_AnalysisRegion_read[1]);
	attr_GaussFit_read[0] = attr_GaussFit_read[1] = attr_GaussFit_read[2] = attr_GaussFit_read[3] = 0;
	attr_FWHM_read[0] = attr_FWHM_read[1] = attr_FWHM_read[2] = 0;
	*attr_GaussFitCenter_read = 0;
	*attr_GaussFitWidth_read = 0;
	*attr_FWHMCenter_read = 0;
	*attr_FWHMWidth_read = 0;
	GaussFit_valid = false;

	Tango::MultiAttribute *attr_list = get_device_attr();

	vector<double> region_write;
	region_write.push_back(x_values[RegionOfInterestStartPix]);
	region_write.push_back(x_values[RegionOfInterestEndPix]);
	Tango::WAttribute &attr_reg = attr_list->get_w_attr_by_name ("RegionOfInterest");
	attr_reg.set_write_value(region_write,2);
	Tango::WAttribute &attr_reg2 = attr_list->get_w_attr_by_name ("AnalysisRegion");
	attr_reg2.set_write_value(region_write,2);
	if(reconnection)
	{
		usleep(500000);
		try
		{
			Tango::WAttribute &attr_reg3 = attr_list->get_w_attr_by_name ("ExposureTime");
			write_ExposureTime(attr_reg3);
		}catch (...){}
		usleep(100000);
		try
		{
			Tango::WAttribute &attr_reg4 = attr_list->get_w_attr_by_name ("SyncMode");
			write_SyncMode(attr_reg4);
		}catch (...){}
		usleep(100000);
	}


	if(nFramesBuffered > 0)
	{
		if(pFrameBuffer != NULL)
		{
			delete [] pFrameBuffer;
			pFrameBuffer = NULL;
		}
		pFrameBuffer = new double[NumPixel * nFramesBuffered];
		memset(pFrameBuffer,0,NumPixel * nFramesBuffered * sizeof(DWORD));
	}


	ifstream  fd;		//startup file
	fd.open(dataReferenceFile.c_str());
	if (!fd)
	{
		stringstream tmp;
		tmp << "Cannot open Data Reference default file:" << strerror(errno) << endl;
		INFO_STREAM << "OceanOptics::OceanOptics(): " << tmp.str();
   	}
	else{
		DEBUG_STREAM << "OceanOptics::OceanOptics(): file opened successfully" << endl;
		string lineread;
		int i=0;
		double val;
		while(std::getline(fd, lineread, '\n'))
		{ // Read line by line
			if(lineread.length() > 0 && i < NumPixel)
			{
				sscanf(lineread.c_str(),"%lf",&val);
				pReferenceData[i++] = val;
			}
		}
		fd.close();
	}

	fd.open(noiseProfileFile.c_str());
	if (!fd)
	{
		stringstream tmp;
		tmp << "Cannot open Noise Profile default file:" << strerror(errno) << endl;
		INFO_STREAM << "OceanOptics::OceanOptics(): " << tmp.str();
   	}
	else{
		DEBUG_STREAM << "OceanOptics::OceanOptics(): file " << noiseProfileFile << " opened successfully" << endl;
		string lineread;
		int i=0;
		double val;
		while(std::getline(fd, lineread, '\n'))
		{ // Read line by line
			if(lineread.length() > 0 && i < NumPixel)
			{
				sscanf(lineread.c_str(),"%lf",&val);
				pNoiseMask_avg[i++] = val;
			}
		}
		fd.close();
	}
}

unsigned int OceanOptics::WaveLenght2Pix(double val)
{
	if(!x_values || NumPixel==0)
		return 0;
	double tmp_val;
	unsigned int tmp_pix, step;
	tmp_pix= (unsigned int)(((double)NumPixel)/2);
	step= (unsigned int)(((double)NumPixel)/2);
	tmp_val = x_values[tmp_pix];
	//DEBUG_STREAM << "OceanOptics::WaveLenght2Pix: ENTERING: val="<<val<<" tmp_val="<<tmp_val<<" tmp_pix="<<tmp_pix<<" step="<<step<<endl;
	if(val <= x_values[0])
		return 0;
	else if(val >= x_values[NumPixel-1])
		return NumPixel-1;
	while((val != tmp_val) && (step > 1))
	{
		step = (unsigned int)((((double)step)/2) + 0.5);	//round
		//DEBUG_STREAM << "OceanOptics::WaveLenght2Pix: LOOPING: val="<<val<<" tmp_val="<<tmp_val<<" tmp_pix="<<tmp_pix<<" step="<<step<<endl;
		if(val > tmp_val)
		{
			if((((int)tmp_pix) + ((int)step)) < (int)NumPixel)
			{
				tmp_pix = tmp_pix + step;
			}
			else
			{
				//DEBUG_STREAM << "OceanOptics::WaveLenght2Pix("<<val<<"): LOOPING: tmp_pix="<<tmp_pix<<"+step="<<step<<" >=NumPixel="<<NumPixel<<endl;
				tmp_pix = NumPixel-1;
				//break;
			}
		}
		else if(val < tmp_val)
		{
			if((((int)tmp_pix) - ((int)step)) >= 0)
			{
				//DEBUG_STREAM << "OceanOptics::WaveLenght2Pix("<<val<<"): LOOPING: tmp_pix="<<tmp_pix<<"-step="<<step<<" ="<<tmp_pix - step<<" new tmp_val=" << x_values[tmp_pix- step] << endl;
				tmp_pix = tmp_pix - step;
			}
			else
			{
				//DEBUG_STREAM << "OceanOptics::WaveLenght2Pix("<<val<<"): LOOPING: tmp_pix="<<tmp_pix<<"-step="<<step<<" <=0"<<endl;
				tmp_pix = 0;
				//break;
			}
		}
		else
		{
			break;
		}
		tmp_val = x_values[tmp_pix];
	}
	if(step == 1)
	{
		//DEBUG_STREAM << "OceanOptics::WaveLenght2Pix: LAST TEST: val="<<val<<" tmp_val="<<tmp_val<<" tmp_pix="<<tmp_pix<<" step="<<step<<endl;
		if(val > tmp_val)
		{
			if((((int)tmp_pix) + ((int)step)) < (int)NumPixel)
			{
				tmp_pix = tmp_pix + step;
			}
			else
			{
				//DEBUG_STREAM << "OceanOptics::WaveLenght2Pix("<<val<<"): LAST TEST: tmp_pix="<<tmp_pix<<"+step="<<step<<" >=NumPixel="<<NumPixel<<endl;
				tmp_pix = NumPixel-1;
				//break;
			}
		}
		else if(val < tmp_val)
		{
			if((((int)tmp_pix) - ((int)step)) >= 0)
			{
				//DEBUG_STREAM << "OceanOptics::WaveLenght2Pix("<<val<<"): LAST TEST: tmp_pix="<<tmp_pix<<"-step="<<step<<" ="<<tmp_pix - step<<" new tmp_val=" << x_values[tmp_pix- step] << endl;
				tmp_pix = tmp_pix - step;
			}
			else
			{
				DEBUG_STREAM << "OceanOptics::WaveLenght2Pix("<<val<<"): LAST TEST: tmp_pix="<<tmp_pix<<"-step="<<step<<" <=0"<<endl;
				tmp_pix = 0;
				//break;
			}
		}
	}
	//DEBUG_STREAM << "OceanOptics::WaveLenght2Pix: EXITING: val="<<val<<" tmp_val="<<tmp_val<<" tmp_pix="<<tmp_pix<<" step="<<step<<endl;
	return (unsigned int)tmp_pix;
}
//==== argin:
// DWORD *data_array: sequence of "num_frames" frames
// int dim_frame: dimensione of one frame
// int x_start_analysis: start of region of analysis within the frame
// int x_end_analysis: end of region of analysis within the frame
// int num_frames: number of frames passed for the average between frames
// int avg_val: number of adjacent values with which calculate the average
//
//==== argout:
// int x_peak: x of the peak
// int y_peak: y of the peak
// int x_half_min: smallest x of the value at half maximum
// int x_half_max: largest x of the value at half maximum

bool OceanOptics::fwhm_calculate(double *data_array, size_t dim_frame, size_t x_start_analysis, size_t x_end_analysis, size_t num_frames, size_t avg_val, size_t &x_peak, double &y_peak, size_t &x_half_min, size_t &x_half_max)
{
	bool ret = true;
	if(num_frames < 1)
		num_frames = 1;
	int dim_frame_analysis = x_end_analysis - x_start_analysis + 1;
	if(avg_val-1 > dim_frame_analysis)
	{
		avg_val = dim_frame_analysis;
	}
	if(avg_val < 1)
		avg_val = 1;
	if((avg_val % 2) != 1)		//has to be odd
		avg_val += 1;

	size_t half_avg_val = (size_t)(floor((double)avg_val)/2);

	double *data_tmp = new double[dim_frame_analysis];
	//memset(data_tmp, 0, sizeof(double)*dim_frame_analysis);
	double y_min=1E17;	//TODO: start from max value
	x_peak = x_start_analysis + avg_val;
	y_peak = -1E17;
	x_half_min=x_start_analysis + avg_val;
	x_half_max=x_end_analysis-(int)(((double)avg_val)/2);

	double inv_num_frame = 1.0/(double)num_frames;
	double inv_avg_val = 1.0/(double)avg_val;
	//cout << "GGG avg_val="<<avg_val<<" looping ["<<x_start_analysis + avg_val<<","<<x_end_analysis - avg_val<<"]" << endl;
	//calculate average and max
	//if(int_num_frames > 1)
	{
		for(size_t i=x_start_analysis + avg_val; i < x_end_analysis - avg_val; i++)
		{
			//average with num_frames before
			for(size_t j=0; j < num_frames; j++)
			{
				if(avg_val == 1)
				{
					data_tmp[i-x_start_analysis] = ((double)data_array[i + j*num_frames])*inv_num_frame;
				}
				else
				{
					double tmp_val=0;
					//average from (size_t)(((double)avg_val)/2) before value i to (size_t)(((double)avg_val)/2) after
					for(size_t k=i-half_avg_val; k<=i+half_avg_val; k++)
					{
						tmp_val += ((double)data_array[k + j*num_frames]);
						//cout << "GGG i="<<i<<" k="<<k<<" data_array["<<k + j*num_frames<<"]="<< ((double)data_array[k + j*num_frames]) << endl;
					}
					data_tmp[i-x_start_analysis] = tmp_val*inv_num_frame*inv_avg_val;
					//cout << "GGG i="<<i<<" data_tmp["<<i-x_start_analysis<<"]="<< data_tmp[i-x_start_analysis] << endl;
				}
			}
			if(data_tmp[i-x_start_analysis] > y_peak)
			{
				y_peak = (data_tmp[i-x_start_analysis]);
				x_peak = i;
				//cout << "GGG FOUND peak: i="<<i<<" data_tmp["<<i-x_start_analysis<<"]="<< data_tmp[i-x_start_analysis] << endl;
			}
			if(data_tmp[i-x_start_analysis] < y_min)
			{
				y_min = (data_tmp[i-x_start_analysis]);
			}
		}

		//now search for x at half maximum
		//exterior half maximum on the left
		//for(int i=x_start_analysis + avg_val; i < x_peak; i++)
		for(int i=x_peak-1; i>=x_start_analysis + avg_val; i--)
		{
			//cout << "GGG i="<<i<<" data_tmp["<<i-x_start_analysis<<"]="<< data_tmp[i-x_start_analysis] <<" y_min="<<y_min<<" y_peak="<<y_peak<<" x_peak="<<x_peak<< endl;
			//if(data_tmp[i-x_start_analysis] >= (double)y_min + ((double)y_peak-y_min)/2)
			if(data_tmp[i-x_start_analysis] < (double)y_min + ((double)y_peak-y_min)/2)
			{
				//cout << "GGG FOUND x_half_min="<<i<< endl;
				//x_half_min = i;
				x_half_min = i+1;
				break;
			}
		}

		//exterior half maximum on the right
		//for(int i=x_end_analysis - avg_val; i > x_peak; i--)
		for(int i = x_peak+1; i <= x_end_analysis - avg_val; i++)
		{
			//cout << "GGG i="<<i<<" data_tmp["<<i-x_start_analysis<<"]="<< data_tmp[i-x_start_analysis] <<" y_min="<<y_min<<" y_peak="<<y_peak<<" x_peak="<<x_peak<< endl;
			//if(data_tmp[i-x_start_analysis] >= (double)y_min + ((double)y_peak-y_min)/2)
			if(data_tmp[i-x_start_analysis] < (double)y_min + ((double)y_peak-y_min)/2)
			{
				//cout << "GGG FOUND x_half_max="<<i<< endl;
				//x_half_max = i;
				x_half_max = i-1;
				break;
			}
		}
	}
	delete [] data_tmp;
	return ret;
}

uint32_t OceanOptics::myntohl(uint32_t val_l)
{
	volatile uint32_t ret_val=val_l;
	uint32_t volatile *pval_l;
	pval_l = (uint32_t volatile *)&ret_val;

#if __BYTE_ORDER == __BIG_ENDIAN		//NOT STANDARD: ASSUMING 	__LITTLE_ENDIAN as the network ordering
	*pval_l = __bswap_32(*pval_l);
#else
	*pval_l = *pval_l;
#endif

	return ret_val;
}

int OceanOptics::connect_device(bool reconnection = false)
{
	int ret=0;

	device_id = -1;
#if _SEABREEZE_API > 1
	spectrometer_id = -1;
	{
		int result=1;
		int number_of_devices;
		long *device_ids;
		int i;
		int test_index;
		int flag;
		int error = 0;
		char nameBuffer[80];

		/* Give the driver a chance to initialize itself */
		sbapi_initialize();

		printf("Probing for devices...\n"); fflush(stdout);
		sbapi_probe_devices();

		printf("Getting device count...\n"); fflush(stdout);
		number_of_devices = sbapi_get_number_of_device_ids();

		printf("Device count is %d\n", number_of_devices);
		if(0 == number_of_devices)
		{
			result = 0;
		}

		printf("Getting device IDs...\n");
		device_ids = (long *)calloc(number_of_devices, sizeof(long));
		number_of_devices = sbapi_get_device_ids(device_ids, number_of_devices);
		printf("Got %d device ID%s.\n", number_of_devices, number_of_devices == 1 ? "" : "s");

		for(i = 0; i < number_of_devices; i++)
		{


			printf("%d: Device 0x%02lX:\n", i, device_ids[i]);
			printf("\tGetting device type...\n");
			flag = sbapi_get_device_type(device_ids[i], &error, nameBuffer, 79);
			printf("\t\tResult is (%d) [%s]\n", flag, sbapi_get_error_string(error));
			if(flag > 0)
			{
				printf("\tDevice type: [%s]\n", nameBuffer);
			}

			/* Open the device */
			printf("\tAttempting to open:\n");
			flag = sbapi_open_device(device_ids[i], &error);
			printf("\t\tResult is (%d) [%s]\n", flag, sbapi_get_error_string(error));
			if(flag != 0)
			{
				continue;
			}

			int error = 0;
			int number_of_serial_numbers;
			long *serial_number_ids = 0;
			int j;
			char buffer[80];

			printf("\tTesting serial number features:\n");

			printf("\t\tGetting number of serial numbers:\n");
			number_of_serial_numbers = sbapi_get_number_of_serial_number_features(device_ids[i], &error);
			printf("\t\t\tResult is %d [%s]\n", number_of_serial_numbers,
			sbapi_get_error_string(error));

			if(0 == number_of_serial_numbers)
			{
				printf("\tNo serial number capabilities found.\n");
				continue;
			}

			serial_number_ids = (long *)calloc(number_of_serial_numbers, sizeof(long));
			printf("\t\tGetting serial number feature IDs...\n");
			number_of_serial_numbers = sbapi_get_serial_number_features(device_ids[i], &error,
					serial_number_ids, number_of_serial_numbers);
			printf("\t\t\tResult is %d [%s]\n", number_of_serial_numbers,
			sbapi_get_error_string(error));

			for(j = 0; j < number_of_serial_numbers; j++)
			{
				printf("\t\t%d: Testing device 0x%02lX, serial number 0x%02lX\n",
						j, device_ids[i], serial_number_ids[j]);

				printf("\t\t\tAttempting to get serial number...\n");
				memset(buffer, (int)0, sizeof(buffer));
				sbapi_get_serial_number(device_ids[i],serial_number_ids[j], &error, buffer, 79);
				printf("\t\t\t\tResult is [%s]\n", sbapi_get_error_string(error));
				if(0 == error)
				{
					printf("\t\t\t\tSerial number: [%s]\n", buffer);
					if(string(buffer) == serialNumber)
					{
						device_id = device_ids[i];
						DEBUG_STREAM << "OceanOptics::OceanOptics(): found serial num '" <<serialNumber<<"' on USB dev ID=" <<device_ids[i];
						break;
					}
				}
				printf("\t\t%d: Finished testing device 0x%02lX, serial number 0x%02lX\n",
						j, device_ids[i], serial_number_ids[j]);
			}
			free(serial_number_ids);
			if(device_id != -1)
				break;
			else
			{
				/* Close the device */
				printf("\tAttempting to close:\n");
				sbapi_close_device(device_ids[i], &error);
				printf("\t\tResult is (%d) [%s]\n", flag, sbapi_get_error_string(error));
			}
		}
		free(device_ids);


		if(device_id != -1)
		{
			int error = 0;
			int length;
			long integration_time;
			int number_of_spectrometers;
			long *spectrometer_ids = 0;
			int i;
			double *doubleBuffer = 0;
			unsigned char *byteBuffer = 0;
			int *edarkBuffer = 0;
			int j;

			printf("\tTesting spectrometer features:\n");

			printf("\t\tGetting number of spectrometers:\n");
			number_of_spectrometers = sbapi_get_number_of_spectrometer_features(device_id, &error);
			printf("\t\t\tResult is %d [%s]\n", number_of_spectrometers,
					sbapi_get_error_string(error));

			if(0 == number_of_spectrometers)
			{
				printf("\tNo spectrometer capabilities found.\n");
				return;
			}

			spectrometer_ids = (long *)calloc(number_of_spectrometers, sizeof(long));
			printf("\t\tGetting spectrometer feature IDs...\n");
			number_of_spectrometers = sbapi_get_spectrometer_features(device_id, &error,
			spectrometer_ids, number_of_spectrometers);
			printf("\t\t\tResult is %d [%s]\n", number_of_spectrometers,
					sbapi_get_error_string(error));
			if(number_of_spectrometers == 1)
			{
				spectrometer_id = spectrometer_ids[0];
			}
			else
			{
				printf("\nNumber of spectrometer features = %d. Only 1 supported! EXITING\n", number_of_spectrometers);
				exit(-1);
			}
		}
	}
	//else
	if(device_id == -1)
#else
    int flag;
    int error;
    char error_str[128];
    char type[16];
    int device_count = 0;
	int i;
	for(i = 0; i < SEABREEZE_MAX_DEVICES; i++)
	{
		printf("Opening spectrometer %d.\n", i);
		flag = seabreeze_open_spectrometer(i, &error);
		seabreeze_get_error_string(error, error_str, 128);
		printf("Result is (%d) [%s]\n", flag, error_str);
		if(0 == flag)
		{
			char serial_number[32];
		    printf("Getting serial number.\n");
		    flag = seabreeze_get_serial_number(i, &error, serial_number, 32);
		    seabreeze_get_error_string(error, error_str, 128);
		    printf("Result is (%d) [%s]\n", flag, error_str);
		    serial_number[31] = '\0';
		    if(flag > 0)
		    {
		        printf("Serial number: [%s]\n", serial_number);
				if(string(serial_number) == serialNumber)
				{
					device_id = i;
					printf("Getting device %d name.\n", i);
					seabreeze_get_model(i, &error, type, sizeof(type));
					seabreeze_get_error_string(error, error_str, 128);
					printf("Result is (%s) [%s]\n", type, error_str);
					DEBUG_STREAM << "OceanOptics::OceanOptics(): found serial num '" <<serialNumber<<"' on USB dev ID=" <<i<<" Type="<<type;
					break;
				}
		    }
	        printf("Closing spectrometer %d.\n", i);
	        flag = seabreeze_close_spectrometer(i, &error);
	        seabreeze_get_error_string(error, error_str, 128);
	        printf("Result is (%d) [%s]\n", flag, error_str);
		}
		else
		{
			break;
		}
	}
	if(device_id == -1)
#endif
	{
		ERROR_STREAM << "OceanOptics::OceanOptics(): error initing CCD!" << endl;
		set_state(Tango::FAULT);
		set_status("Communication error");
		//exit(0);
		return -1;
	}
	//set temporary state on to allow write attr in init_data
	set_state(Tango::ON);
	init_data(reconnection);

	if(device_id != -1)
	{
		if(live_mode)
		{
			set_state(Tango::RUNNING);
			set_status("Continuous Mode");
		}
		else
		{
			set_state(Tango::ON);
			set_status("Ready");
		}
	}

	return ret;
}

int OceanOptics::disconnect_device()
{
#if _SEABREEZE_API > 1
	if(device_id != -1)
	{
		int flag;
		int error;
		printf("\tAttempting to close:\n");
		sbapi_close_device(device_id, &error);
		printf("\t\tResult is (%d) [%s]\n", flag, sbapi_get_error_string(error));
	}
	/* Clean up memory allocated by the driver */
	sbapi_shutdown();
#else
	if(device_id != -1)
	{
		int flag;
		int error;
		char error_str[128];
		printf("Closing spectrometer %d.\n", device_id);
		flag = seabreeze_close_spectrometer(device_id, &error);
		seabreeze_get_error_string(error, error_str, 128);
		printf("Result is (%d) [%s]\n", flag, error_str);
	}
    seabreeze_shutdown();
#endif
}

//+------------------------------------------------------------------
/**
 *	method:	PLCintll::string_explode
 * Explode a string
*/
//+------------------------------------------------------------------
void OceanOptics::string_explode(string str, string separator, vector<string>* results)
{
	size_t found;

	found = str.find_first_of(separator);
	while(found != string::npos) {
		if(found > 0) {
			results->push_back(str.substr(0,found));
		}
		str = str.substr(found+1);
		found = str.find_first_of(separator);
	}
	if(str.length() > 0) {
		results->push_back(str);
	}

}

int OceanOptics::load_intensity_cal_file(string filename)
{
	string line;
	ifstream infile (filename.c_str(), ios_base::in);
	while (getline(infile, line, '\n'))
	{
		vector<string> results;
		string_explode(line, "\t", &results);
		if(results.size() == 3)
		{
			int ind;
			float wl, cal;
			int ret = sscanf(line.c_str(), "%d\t%f\t%f", &ind, &wl, &cal);
			DEBUG_STREAM <<__func__ << ":Parsed from " << filename << " ret=" << ret << " cal[" << ind << "]=" << cal;
			if(ret == 3 && ind >=1 && ind <= NumPixel)
			{
				attr_IntensityCalibration_read[ind-1] = cal;
			}
		}
	}

	infile.close();
	//DEBUG_STREAM <<__func__ << ": Read from " << filename << " " << tmp.size << " lines or " << tmp.size_d << " reals";
}

int OceanOptics::save_intensity_cal_file(string filename)
{
	string line;
	ofstream outfile (filename.c_str(), ios_base::out | ios_base::trunc);
	for(size_t i=0; i<NumPixel; i++)
	{
		outfile << i+1 << "\t" << x_values[i] << "\t" << attr_IntensityCalibration_read[i] << endl;
	}

	outfile.close();
}

#if 0
//f(x) = base + amp*exp(-0.5 * ((x-center)/width)^2)
int gauss_f (const gsl_vector * x, void *data, gsl_vector * f)
{
	size_t n = ((struct internal_data *)data)->n;
	double *y = ((struct internal_data *)data)->y;
	double *sigma = ((struct internal_data *) data)->sigma;
	double *x_values = ((struct internal_data *) data)->x_values;

#ifdef _DEBUG_FIT
	printf ("gauss_f: xval[0]= % 15.8f xval[%u]=% 15.8f\n",x_values[0],n-1,x_values[n-1]);
#endif
	/*double A = gsl_vector_get (x, 0);
	double lambda = gsl_vector_get (x, 1);
	double b = gsl_vector_get (x, 2);*/

	double base = gsl_vector_get (x, 0);
	double amp = gsl_vector_get (x, 1);
	double center = gsl_vector_get (x, 2);
	double width = gsl_vector_get (x, 3);

	//TODO: add here bounds to center??
	//TEST!!!!!!!!!!!!!!!!!!!!!!!!!!
/*	if(center < x_values[0])
	{
		center = x_values[1];
		gsl_vector_set(x,2,center);
	}
	if(center > x_values[n-1])
	{
		center = x_values[n-2];
		gsl_vector_set(x,2,center);
	}
	if(amp < 0)
	{
		amp = 1;
		gsl_vector_set(x,1,amp);
	}*/
	//END TEST!!!!!!!!!!!!!!!!!!!!!!

	size_t i;

	for (i = 0; i < n; i++)
	{
		//Model Yi = base + amp * exp((-0.5) * ((i-center)/width)^2)
		double Yi = base + amp * exp((-0.5) * pow((x_values[i]-center)/width,2));
		gsl_vector_set (f, i, (Yi - y[i])/sigma[i]);
	}

  return GSL_SUCCESS;
}
int gauss_df (const gsl_vector * x, void *data, gsl_matrix * J)
{
	size_t n = ((struct internal_data *)data)->n;
	double *sigma = ((struct internal_data *) data)->sigma;
	double *x_values = ((struct internal_data *) data)->x_values;

#ifdef _DEBUG_FIT
	printf ("gauss_df: xval[0]= % 15.8f xval[%u]=% 15.8f\n",x_values[0],n-1,x_values[n-1]);
#endif

	/*double A = gsl_vector_get (x, 0);
	double lambda = gsl_vector_get (x, 1);
	double b = gsl_vector_get (x, 2);*/

	double base = gsl_vector_get (x, 0);
	double amp = gsl_vector_get (x, 1);
	double center = gsl_vector_get (x, 2);
	double width = gsl_vector_get (x, 3);

	//TODO: add here bounds to center??
	//TEST!!!!!!!!!!!!!!!!!!!!!!!!!!
	/*if(center < x_values[0])
	{
		center = x_values[1];
		gsl_vector_set(x,2,center);
	}
	if(center > x_values[n-1])
	{
		center = x_values[n-2];
		gsl_vector_set(x,2,center);
	}
	if(amp < 0)
	{
		amp = 1;
		gsl_vector_set(x,1,amp);
	}*/
	//END TEST!!!!!!!!!!!!!!!!!!!!!!

	size_t i;

	for (i = 0; i < n; i++)
    {
		double s = sigma[i];

		//the four partial derivative of
		//base + amp * exp((-0.5) * pow((i-center)/width,2))
		double d_base = 1;
		double d_amp = exp((-0.5) * pow((x_values[i]-center)/width,2));
		double d_center = (amp * exp((-0.5) * pow((x_values[i]-center)/width,2))) * ((x_values[i]-center)/pow(width,2));
		double d_width = (amp * exp((-0.5) * pow((x_values[i]-center)/width,2))) * (pow(x_values[i]-center,2)/pow(width,3));
		//populate the Jacobian
		gsl_matrix_set (J, i, 0, d_base/s); 
		gsl_matrix_set (J, i, 1, d_amp/s);
		gsl_matrix_set (J, i, 2, d_center/s);
		gsl_matrix_set (J, i, 3, d_width/s);

    }
  return GSL_SUCCESS;
}
int gauss_fdf (const gsl_vector * x, void *data, gsl_vector * f, gsl_matrix * J)
{
  gauss_f (x, data, f);
  gauss_df (x, data, J);

  return GSL_SUCCESS;
}
#endif


/*----- PROTECTED REGION END -----*/	//	OceanOptics::namespace_ending
} //	namespace
