/*----- PROTECTED REGION ID(OceanOptics.h) ENABLED START -----*/
//=============================================================================
//
// file :        OceanOptics.h
//
// description : Include file for the OceanOptics class
//
// project :     SolarLaserSystems S100
//
// This file is part of Tango device class.
// 
// Tango is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Tango is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Tango.  If not, see <http://www.gnu.org/licenses/>.
// 
// $Author: graziano $
//
// $Revision: 1.11 $
// $Date: 2019-04-23 13:32:28 $
//
// $HeadURL:  $
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#ifndef OceanOptics_H
#define OceanOptics_H

#define _SEABREEZE_API	1

#include <tango.h>
#include <stdint.h>
//#include "UsbCDD.h"
//#include "GaussFit.h"
#if _SEABREEZE_API > 1
#include "api/seabreezeapi/SeaBreezeAPI.h"
#else
#include "api/SeaBreezeWrapper.h"
#endif
#include "Fit.h"
//#include "buffer.h"

#include "tbuffer.hpp"

#define DWORD uint32_t
#define MAX_GAUSS_ITERATION     25
#define MAX_CONFITEOR_ITERATION 20
#define NUM_GAUSS_PARAMETER     4
#define AUTO_EXPOSURE_MEAN  4
#define SIGMA_TO_FWHM_FACTOR    0.4246609

#define NUM_BUF		5
#define BUF_SIZE			10000
#define BUF_ID_FWHM_CENTER		0
#define BUF_ID_FWHM_WIDTH		1
#define BUF_ID_GAUSS_CENTER		2
#define BUF_ID_GAUSS_WIDTH		3
#define BUF_ID_AREA				4


// The synchronization mode
// Without synchronization.
#define TG_SYNCHR_NONE		0
// In the beginning of the first accumulation the positive
// pulse of synchronization is formed.
#define TG_SYNCHR_CONTR		1
// Clock pulse is formed in the beginning of each accumulation.
#define TG_SYNCHR_CONTR_FRS	2   
// One pulse of synchronization is formed on all time of registration.
// A pulse of negative polarity.
#define TG_SYNCHR_CONTR_NEG	3   
                          
// The beginning of the first accumulation is adhered to growing
// front of external clock pulse.
//All other accumulation occur so quickly as it is possible.
// In a limit -- without the misses.
#define TG_SYNCHR_EXT		4

// The beginning of each accumulation is adhered to growing front of clock pulse.
// How much accumulation, so much clock pulses are expected.
#define TG_SYNCHR_EXT_FRS	5 

class readthread;

#if 0
//f(x) = base + amp*exp(-0.5 * ((x-center)/width)^2)
int gauss_f (const gsl_vector * x, void *data, gsl_vector * f);

int gauss_df (const gsl_vector * x, void *data, gsl_matrix * J);

int gauss_fdf (const gsl_vector * x, void *data, gsl_vector * f, gsl_matrix * J);
#endif

/*----- PROTECTED REGION END -----*/	//	OceanOptics.h

/**
 *  OceanOptics class description:
 *    Tango device server for OceanOptics compact wide-range spectrometer
 */

namespace OceanOptics_ns
{
/*----- PROTECTED REGION ID(OceanOptics::Additional Class Declarations) ENABLED START -----*/

//	Additional Class Declarations

/*----- PROTECTED REGION END -----*/	//	OceanOptics::Additional Class Declarations

class OceanOptics : public TANGO_BASE_CLASS
{

/*----- PROTECTED REGION ID(OceanOptics::Data Members) ENABLED START -----*/

//	Add your own data members
public:

	Tango::DevLong attr_FitType_write;
	Tango::DevLong attr_FWHMAvg_write;
	Tango::DevLong attr_NoiseLevel_write;
	Tango::DevBoolean attr_AutoExposure_write;
	Tango::DevDouble attr_ExposureTime_write;
#if _SEABREEZE_API > 1
	long device_id;
	long spectrometer_id;
#else
	int device_id;
#endif
	uint32_t myntohl(uint32_t val_l);
	//struct _buffer *buf[NUM_BUF];
	tbuffer::buffer_t<double> tbuf[NUM_BUF];
	tbuffer::buffer_t<tbuffer::_data_array<double> > tbuf_array;//GetData
	tbuffer::buffer_t<tbuffer::_data_array<double> > tbuf_array_w;//GetWave
	tbuffer::buffer_t<tbuffer::_data_array<double> > tbuf_array_l;//GetLambda

/*----- PROTECTED REGION END -----*/	//	OceanOptics::Data Members

//	Device property data members
public:
	//	DeviceId:	USB Id of the device
	Tango::DevLong	deviceId;
	//	SerialNumber:	Serial number of the device
	string	serialNumber;
	//	NFramesBuffered:	number of frames buffered when in live mode
	Tango::DevLong	nFramesBuffered;
	//	DefaultStartRegion:	Beginning of the region of interest
	Tango::DevDouble	defaultStartRegion;
	//	DefaultEndRegion:	End of the region of interest
	Tango::DevDouble	defaultEndRegion;
	//	SleepReading:	Sleep time between reading 2 frames (ms)
	Tango::DevLong	sleepReading;
	//	DefaultGaussRef:	Default Gauss Reference
	vector<Tango::DevDouble>	defaultGaussRef;
	//	DataReferenceFile:	Default Data Reference File
	string	dataReferenceFile;
	//	PreferFWHM:	Choose FWHM instead og GaussianFit if both enabled
	Tango::DevBoolean	preferFWHM;
	//	AutoExposureRange:	AutoExposure Range
	vector<Tango::DevDouble>	autoExposureRange;
	//	AutoExposureStep:	AutoExposure Step
	Tango::DevDouble	autoExposureStep;
	//	AutoExposureAmpRange:	AutoExposure Range
	vector<Tango::DevDouble>	autoExposureAmpRange;
	//	IntensityCalibrationFile:	
	string	intensityCalibrationFile;
	//	NoiseProfileFile:	Default Noise Profile File
	string	noiseProfileFile;
	//	WavelengthAttr:	device name
	//  attribute name
	vector<string>	wavelengthAttr;
	//	AreaInAnalysisRegion:	use AnalysisRegion for area instead of AreaRegion
	Tango::DevBoolean	areaInAnalysisRegion;
	//	WidthCorrection:	widht_in^2 + widthcorrection^2 = widht_out^2
	Tango::DevDouble	widthCorrection;
	//	AnalysisRegion:	AnalysisRegion
	vector<Tango::DevDouble>	analysisRegion;
	//	RegionOfInterest:	RegionOfInterest
	vector<Tango::DevDouble>	regionOfInterest;
	//	AutoStart:	AutoStart
	Tango::DevBoolean	autoStart;
	//	ExposureTime:	ExposureTime
	Tango::DevDouble	exposureTime;
	//	SyncMode:	SyncMode
	Tango::DevLong	syncMode;

//	Attribute data members
public:
	Tango::DevDouble	*attr_ExposureTime_read;
	Tango::DevBoolean	*attr_IncrSensitivity_read;
	Tango::DevLong	*attr_SyncMode_read;
	Tango::DevBoolean	*attr_NoiseSubtraction_read;
	Tango::DevBoolean	*attr_GaussFitEnable_read;
	Tango::DevBoolean	*attr_FWHMEnable_read;
	Tango::DevDouble	*attr_GaussFitCenter_read;
	Tango::DevDouble	*attr_GaussFitWidth_read;
	Tango::DevDouble	*attr_FWHMCenter_read;
	Tango::DevDouble	*attr_FWHMWidth_read;
	Tango::DevDouble	*attr_Center_read;
	Tango::DevDouble	*attr_Width_read;
	Tango::DevLong	*attr_FitType_read;
	Tango::DevLong	*attr_FWHMAvg_read;
	Tango::DevLong	*attr_NoiseLevel_read;
	Tango::DevString	*attr_DataReferenceFile_read;
	Tango::DevBoolean	*attr_AutoExposure_read;
	Tango::DevDouble	*attr_LightSourceIntensity_read;
	Tango::DevBoolean	*attr_RtMode_read;
	Tango::DevBoolean	*attr_LightSourceEnable_read;
	Tango::DevBoolean	*attr_IntensityCalibrationEnable_read;
	Tango::DevDouble	*attr_Area_read;
	Tango::DevDouble	*attr_AreaRegion_read;
	Tango::DevBoolean	*attr_AreaEnable_read;
	Tango::DevBoolean	*attr_WidhtCorrectionEnable_read;
	Tango::DevLong	*attr_NumPixels_read;
	Tango::DevDouble	*attr_Data_read;
	Tango::DevDouble	*attr_RegionOfInterest_read;
	Tango::DevDouble	*attr_AnalysisRegion_read;
	Tango::DevDouble	*attr_GaussFit_read;
	Tango::DevDouble	*attr_FWHM_read;
	Tango::DevDouble	*attr_GaussReference_read;
	Tango::DevDouble	*attr_DataReference_read;
	Tango::DevDouble	*attr_IntensityCalibration_read;
	Tango::DevDouble	*attr_NoiseProfile_read;
	Tango::DevDouble	*attr_Wave_read;
	Tango::DevDouble	*attr_Lambda_read;

//	Constructors and destructors
public:
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device Name
	 */
	OceanOptics(Tango::DeviceClass *cl,string &s);
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device Name
	 */
	OceanOptics(Tango::DeviceClass *cl,const char *s);
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device name
	 *	@param d	Device description.
	 */
	OceanOptics(Tango::DeviceClass *cl,const char *s,const char *d);
	/**
	 * The device object destructor.
	 */
	~OceanOptics() {delete_device();};


//	Miscellaneous methods
public:
	/*
	 *	will be called at device destruction or at init command.
	 */
	void delete_device();
	/*
	 *	Initialize the device
	 */
	virtual void init_device();
	/*
	 *	Read the device properties from database
	 */
	void get_device_property();
	/*
	 *	Always executed method before execution command method.
	 */
	virtual void always_executed_hook();


//	Attribute methods
public:
	//--------------------------------------------------------
	/*
	 *	Method      : OceanOptics::read_attr_hardware()
	 *	Description : Hardware acquisition for attributes.
	 */
	//--------------------------------------------------------
	virtual void read_attr_hardware(vector<long> &attr_list);
	//--------------------------------------------------------
	/*
	 *	Method      : OceanOptics::write_attr_hardware()
	 *	Description : Hardware writing for attributes.
	 */
	//--------------------------------------------------------
	virtual void write_attr_hardware(vector<long> &attr_list);

/**
 *	Attribute ExposureTime related methods
 *	Description: Exposure time in milliseconds
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_ExposureTime(Tango::Attribute &attr);
	virtual void write_ExposureTime(Tango::WAttribute &attr);
	virtual bool is_ExposureTime_allowed(Tango::AttReqType type);
/**
 *	Attribute IncrSensitivity related methods
 *	Description: false: use standard sensitivity\ntrue: use increased sensitivity
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_IncrSensitivity(Tango::Attribute &attr);
	virtual void write_IncrSensitivity(Tango::WAttribute &attr);
	virtual bool is_IncrSensitivity_allowed(Tango::AttReqType type);
/**
 *	Attribute SyncMode related methods
 *	Description: 0: The camera operates without synchronization.
 *               Spectra are read out periodically in accordance with NumReadouts attribute.
 *               
 *               1: Readout is performed periodically in accordance with NumReadouts attribute.
 *               At the beginning of the first readout a positive sync pulse is formed.
 *               
 *               2: A sync pulse is formed at beginning of each readout.
 *               
 *               3: One sync pulse of negative polarity is generated for the entire readout time.
 *               
 *               4: The camera will wait for one sync pulse.
 *               The first readout will start at the rising edge of an external sync pulse.
 *               All other readouts will be performed as soon as possible.
 *               No blanks within limit
 *               
 *               5: Each readout starts at the rising edge of an external sync pulse.
 *               The camera will wait fo as many sync pulses as much readouts are set.
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
	virtual void read_SyncMode(Tango::Attribute &attr);
	virtual void write_SyncMode(Tango::WAttribute &attr);
	virtual bool is_SyncMode_allowed(Tango::AttReqType type);
/**
 *	Attribute NoiseSubtraction related methods
 *	Description: Enable/Disable subtraction of background noise profile \nacquired with the AcquireNoise command
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_NoiseSubtraction(Tango::Attribute &attr);
	virtual void write_NoiseSubtraction(Tango::WAttribute &attr);
	virtual bool is_NoiseSubtraction_allowed(Tango::AttReqType type);
/**
 *	Attribute GaussFitEnable related methods
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_GaussFitEnable(Tango::Attribute &attr);
	virtual void write_GaussFitEnable(Tango::WAttribute &attr);
	virtual bool is_GaussFitEnable_allowed(Tango::AttReqType type);
/**
 *	Attribute FWHMEnable related methods
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_FWHMEnable(Tango::Attribute &attr);
	virtual void write_FWHMEnable(Tango::WAttribute &attr);
	virtual bool is_FWHMEnable_allowed(Tango::AttReqType type);
/**
 *	Attribute GaussFitCenter related methods
 *	Description: wavelength of peak of Gaussian Fit
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_GaussFitCenter(Tango::Attribute &attr);
	virtual bool is_GaussFitCenter_allowed(Tango::AttReqType type);
/**
 *	Attribute GaussFitWidth related methods
 *	Description: width of Gaussian Fit
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_GaussFitWidth(Tango::Attribute &attr);
	virtual bool is_GaussFitWidth_allowed(Tango::AttReqType type);
/**
 *	Attribute FWHMCenter related methods
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_FWHMCenter(Tango::Attribute &attr);
	virtual bool is_FWHMCenter_allowed(Tango::AttReqType type);
/**
 *	Attribute FWHMWidth related methods
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_FWHMWidth(Tango::Attribute &attr);
	virtual bool is_FWHMWidth_allowed(Tango::AttReqType type);
/**
 *	Attribute Center related methods
 *	Description: wavelength of peak of FWHM or of Gaussian Fit
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_Center(Tango::Attribute &attr);
	virtual bool is_Center_allowed(Tango::AttReqType type);
/**
 *	Attribute Width related methods
 *	Description: width of FWHM or of Gaussian Fit
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_Width(Tango::Attribute &attr);
	virtual bool is_Width_allowed(Tango::AttReqType type);
/**
 *	Attribute FitType related methods
 *	Description: 1=Gaussian 2=Gaussian Asymm 3=Confiteor
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
	virtual void read_FitType(Tango::Attribute &attr);
	virtual void write_FitType(Tango::WAttribute &attr);
	virtual bool is_FitType_allowed(Tango::AttReqType type);
/**
 *	Attribute FWHMAvg related methods
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
	virtual void read_FWHMAvg(Tango::Attribute &attr);
	virtual void write_FWHMAvg(Tango::WAttribute &attr);
	virtual bool is_FWHMAvg_allowed(Tango::AttReqType type);
/**
 *	Attribute NoiseLevel related methods
 *	Description: 
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Scalar
 */
	virtual void read_NoiseLevel(Tango::Attribute &attr);
	virtual void write_NoiseLevel(Tango::WAttribute &attr);
	virtual bool is_NoiseLevel_allowed(Tango::AttReqType type);
/**
 *	Attribute DataReferenceFile related methods
 *	Description: 
 *
 *	Data type:	Tango::DevString
 *	Attr type:	Scalar
 */
	virtual void read_DataReferenceFile(Tango::Attribute &attr);
	virtual void write_DataReferenceFile(Tango::WAttribute &attr);
	virtual bool is_DataReferenceFile_allowed(Tango::AttReqType type);
/**
 *	Attribute AutoExposure related methods
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_AutoExposure(Tango::Attribute &attr);
	virtual void write_AutoExposure(Tango::WAttribute &attr);
	virtual bool is_AutoExposure_allowed(Tango::AttReqType type);
/**
 *	Attribute LightSourceIntensity related methods
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_LightSourceIntensity(Tango::Attribute &attr);
	virtual void write_LightSourceIntensity(Tango::WAttribute &attr);
	virtual bool is_LightSourceIntensity_allowed(Tango::AttReqType type);
/**
 *	Attribute RtMode related methods
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_RtMode(Tango::Attribute &attr);
	virtual void write_RtMode(Tango::WAttribute &attr);
	virtual bool is_RtMode_allowed(Tango::AttReqType type);
/**
 *	Attribute LightSourceEnable related methods
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_LightSourceEnable(Tango::Attribute &attr);
	virtual void write_LightSourceEnable(Tango::WAttribute &attr);
	virtual bool is_LightSourceEnable_allowed(Tango::AttReqType type);
/**
 *	Attribute IntensityCalibrationEnable related methods
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_IntensityCalibrationEnable(Tango::Attribute &attr);
	virtual void write_IntensityCalibrationEnable(Tango::WAttribute &attr);
	virtual bool is_IntensityCalibrationEnable_allowed(Tango::AttReqType type);
/**
 *	Attribute Area related methods
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_Area(Tango::Attribute &attr);
	virtual bool is_Area_allowed(Tango::AttReqType type);
/**
 *	Attribute AreaRegion related methods
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_AreaRegion(Tango::Attribute &attr);
	virtual void write_AreaRegion(Tango::WAttribute &attr);
	virtual bool is_AreaRegion_allowed(Tango::AttReqType type);
/**
 *	Attribute AreaEnable related methods
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_AreaEnable(Tango::Attribute &attr);
	virtual void write_AreaEnable(Tango::WAttribute &attr);
	virtual bool is_AreaEnable_allowed(Tango::AttReqType type);
/**
 *	Attribute WidhtCorrectionEnable related methods
 *	Description: enable widht correction
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_WidhtCorrectionEnable(Tango::Attribute &attr);
	virtual void write_WidhtCorrectionEnable(Tango::WAttribute &attr);
	virtual bool is_WidhtCorrectionEnable_allowed(Tango::AttReqType type);
/**
 *	Attribute NumPixels related methods
 *	Description: long[0]: number of pixels horizontally (number of CCD-array columns)\nlong[1]: number of pixels vertically (number of CCD-array lines)
 *
 *	Data type:	Tango::DevLong
 *	Attr type:	Spectrum max = 2
 */
	virtual void read_NumPixels(Tango::Attribute &attr);
	virtual bool is_NumPixels_allowed(Tango::AttReqType type);
/**
 *	Attribute Data related methods
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 5000
 */
	virtual void read_Data(Tango::Attribute &attr);
	virtual bool is_Data_allowed(Tango::AttReqType type);
/**
 *	Attribute RegionOfInterest related methods
 *	Description: Region of interest returned by the Data attribute\ndouble[0]=start of the region\ndouble[1]=end of the region\n\nValues are in nm according with the conversion factor in x_values\nData attribute property
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 2
 */
	virtual void read_RegionOfInterest(Tango::Attribute &attr);
	virtual void write_RegionOfInterest(Tango::WAttribute &attr);
	virtual bool is_RegionOfInterest_allowed(Tango::AttReqType type);
/**
 *	Attribute AnalysisRegion related methods
 *	Description: Region used for gaussian and fwhm analysis\ndouble[0]=start of the region\ndouble[1]=end of the region\n\nValues are in nm according with the conversion factor in x_values\nData attribute property
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 2
 */
	virtual void read_AnalysisRegion(Tango::Attribute &attr);
	virtual void write_AnalysisRegion(Tango::WAttribute &attr);
	virtual bool is_AnalysisRegion_allowed(Tango::AttReqType type);
/**
 *	Attribute GaussFit related methods
 *	Description: double[0]= wavelength of peak of Gaussian Fit\ndouble[1]= width of Gaussian Fit
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4
 */
	virtual void read_GaussFit(Tango::Attribute &attr);
	virtual bool is_GaussFit_allowed(Tango::AttReqType type);
/**
 *	Attribute FWHM related methods
 *	Description: double[0]= wavelength of peak of FWHM\ndouble[1]= width of FWHM
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 3
 */
	virtual void read_FWHM(Tango::Attribute &attr);
	virtual bool is_FWHM_allowed(Tango::AttReqType type);
/**
 *	Attribute GaussReference related methods
 *	Description: Gaussian of reference params
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 4
 */
	virtual void read_GaussReference(Tango::Attribute &attr);
	virtual void write_GaussReference(Tango::WAttribute &attr);
	virtual bool is_GaussReference_allowed(Tango::AttReqType type);
/**
 *	Attribute DataReference related methods
 *	Description: A reference curve
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 5000
 */
	virtual void read_DataReference(Tango::Attribute &attr);
	virtual void write_DataReference(Tango::WAttribute &attr);
	virtual bool is_DataReference_allowed(Tango::AttReqType type);
/**
 *	Attribute IntensityCalibration related methods
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 10000
 */
	virtual void read_IntensityCalibration(Tango::Attribute &attr);
	virtual void write_IntensityCalibration(Tango::WAttribute &attr);
	virtual bool is_IntensityCalibration_allowed(Tango::AttReqType type);
/**
 *	Attribute NoiseProfile related methods
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 5000
 */
	virtual void read_NoiseProfile(Tango::Attribute &attr);
	virtual bool is_NoiseProfile_allowed(Tango::AttReqType type);
/**
 *	Attribute Wave related methods
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 5000
 */
	virtual void read_Wave(Tango::Attribute &attr);
	virtual bool is_Wave_allowed(Tango::AttReqType type);
/**
 *	Attribute Lambda related methods
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Spectrum max = 5000
 */
	virtual void read_Lambda(Tango::Attribute &attr);
	virtual bool is_Lambda_allowed(Tango::AttReqType type);


	//--------------------------------------------------------
	/**
	 *	Method      : OceanOptics::add_dynamic_attributes()
	 *	Description : Add dynamic attributes if any.
	 */
	//--------------------------------------------------------
	void add_dynamic_attributes();




//	Command related methods
public:
	/**
	 *	Command GetVersion related method
	 *	Description: Returns serial number
	 *
	 *	@returns serial number
	 */
	virtual Tango::DevString get_version();
	virtual bool is_GetVersion_allowed(const CORBA::Any &any);
	/**
	 *	Command Stop related method
	 *	Description: Reset the CCD:
	 *               stops the current readout,
	 *               initiates the internal DLL variables,
	 *               stops data transfer
	 *
	 */
	virtual void stop();
	virtual bool is_Stop_allowed(const CORBA::Any &any);
	/**
	 *	Command StartSingle related method
	 *	Description: Start single (NumReadout frames) acquisition fron the CCD
	 *
	 */
	virtual void start_single();
	virtual bool is_StartSingle_allowed(const CORBA::Any &any);
	/**
	 *	Command StartContinuous related method
	 *	Description: Start acquisition in Continuous mode
	 *
	 */
	virtual void start_continuous();
	virtual bool is_StartContinuous_allowed(const CORBA::Any &any);
	/**
	 *	Command GetLastData related method
	 *	Description: returns the last argin frames saved
	 *
	 *	@param argin number of frames to return
	 *	@returns 
	 */
	virtual Tango::DevVarDoubleArray *get_last_data(Tango::DevLong argin);
	virtual bool is_GetLastData_allowed(const CORBA::Any &any);
	/**
	 *	Command AcquireNoise related method
	 *	Description: Acquire argin frames of background noise
	 *
	 *	@param argin Number of frames to acquire as noise
	 */
	virtual void acquire_noise(Tango::DevLong argin);
	virtual bool is_AcquireNoise_allowed(const CORBA::Any &any);
	/**
	 *	Command SaveReferenceData related method
	 *	Description: Save current curve as reference curve
	 *
	 */
	virtual void save_reference_data();
	virtual bool is_SaveReferenceData_allowed(const CORBA::Any &any);
	/**
	 *	Command OpenShutter related method
	 *	Description: 
	 *
	 */
	virtual void open_shutter();
	virtual bool is_OpenShutter_allowed(const CORBA::Any &any);
	/**
	 *	Command CloseShutter related method
	 *	Description: 
	 *
	 */
	virtual void close_shutter();
	virtual bool is_CloseShutter_allowed(const CORBA::Any &any);
	/**
	 *	Command GetData related method
	 *	Description: 
	 *
	 *	@param argin 
	 *	@returns 
	 */
	virtual Tango::DevVarDoubleArray *get_data(const Tango::DevVarLongArray *argin);
	virtual bool is_GetData_allowed(const CORBA::Any &any);
	/**
	 *	Command GetGaussFitCenter related method
	 *	Description: 
	 *
	 *	@param argin 
	 *	@returns 
	 */
	virtual Tango::DevVarDoubleArray *get_gauss_fit_center(const Tango::DevVarLongArray *argin);
	virtual bool is_GetGaussFitCenter_allowed(const CORBA::Any &any);
	/**
	 *	Command GetGaussFitWidth related method
	 *	Description: 
	 *
	 *	@param argin 
	 *	@returns 
	 */
	virtual Tango::DevVarDoubleArray *get_gauss_fit_width(const Tango::DevVarLongArray *argin);
	virtual bool is_GetGaussFitWidth_allowed(const CORBA::Any &any);
	/**
	 *	Command GetFWHMCenter related method
	 *	Description: 
	 *
	 *	@param argin 
	 *	@returns 
	 */
	virtual Tango::DevVarDoubleArray *get_fwhmcenter(const Tango::DevVarLongArray *argin);
	virtual bool is_GetFWHMCenter_allowed(const CORBA::Any &any);
	/**
	 *	Command GetFWHMWidth related method
	 *	Description: 
	 *
	 *	@param argin 
	 *	@returns 
	 */
	virtual Tango::DevVarDoubleArray *get_fwhmwidth(const Tango::DevVarLongArray *argin);
	virtual bool is_GetFWHMWidth_allowed(const CORBA::Any &any);
	/**
	 *	Command ExportWavelength related method
	 *	Description: 
	 *
	 */
	virtual void export_wavelength();
	virtual bool is_ExportWavelength_allowed(const CORBA::Any &any);
	/**
	 *	Command GetArea related method
	 *	Description: 
	 *
	 *	@param argin 
	 *	@returns 
	 */
	virtual Tango::DevVarDoubleArray *get_area(const Tango::DevVarLongArray *argin);
	virtual bool is_GetArea_allowed(const CORBA::Any &any);
	/**
	 *	Command GetWave related method
	 *	Description: 
	 *
	 *	@param argin 
	 *	@returns 
	 */
	virtual Tango::DevVarDoubleArray *get_wave(const Tango::DevVarLongArray *argin);
	virtual bool is_GetWave_allowed(const CORBA::Any &any);
	/**
	 *	Command GetLambda related method
	 *	Description: 
	 *
	 *	@param argin 
	 *	@returns 
	 */
	virtual Tango::DevVarDoubleArray *get_lambda(const Tango::DevVarLongArray *argin);
	virtual bool is_GetLambda_allowed(const CORBA::Any &any);


	//--------------------------------------------------------
	/**
	 *	Method      : OceanOptics::add_dynamic_commands()
	 *	Description : Add dynamic commands if any.
	 */
	//--------------------------------------------------------
	void add_dynamic_commands();

/*----- PROTECTED REGION ID(OceanOptics::Additional Method prototypes) ENABLED START -----*/

//	Additional Method prototypes
public:
	bool abortflag;
	bool live_mode;
	bool init_live_mode;
	bool measure_mode;
	bool init_measure_mode;
	bool acquire_noise_mode;
	bool init_acquire_noise_mode;

	int noise_frames;
	int noise_frames_counter;
	int noise_level;
	timespec last_read;

	//int data_counter_live;
	//int data_counter_meas;
//	CUsbCDD m_usbCCD;
//	GaussFit *m_gaussfit;//
	/*LibFit_ns::*/Fit *m_gaussfit;//
	/*LibFit_ns::*/Fit *m_asyngaussfit;//
	/*LibFit_ns::*/Fit *m_confiteorfit;//
	bool GaussFit_valid;

	//DWORD *pMeasData;
	double *pLiveData;
	double *pLiveData_avg;
	double *pNoiseMask_avg;
	double *pFrameBuffer;		//bi-dimensional array
	double *pReferenceData;
	unsigned int FrameBufferIndex;
	unsigned int NumPixel;
	Tango::DevDouble *x_values;
	unsigned int RegionOfInterestStartPix;
	unsigned int RegionOfInterestEndPix;
	unsigned int AnalysisRegionStartPix;
	unsigned int AnalysisRegionEndPix;
	void init_data(bool reconnection);
	unsigned int WaveLenght2Pix(double val);
	bool fwhm_calculate(double */*restrict*/ data_array, size_t dim_frame, size_t x_start_analysis, size_t x_end_analisys, size_t num_frames, size_t avg_val, size_t &x_peak, double &y_peak, size_t &x_half_min, size_t &x_half_max);
	int y_max_mean;
	int exposures[AUTO_EXPOSURE_MEAN];
	int exposure_idx;
	int connect_device(bool reconnection);
	int disconnect_device();
	void string_explode(string str, string separator, vector<string>* results);
	int load_intensity_cal_file(string filename);
	int save_intensity_cal_file(string filename);

protected :	
	readthread *loop;
	unsigned int properties;

	char   attr_strchar[1024];
	Tango::DevString attr_tmp_ptr;

/*----- PROTECTED REGION END -----*/	//	OceanOptics::Additional Method prototypes
};

/*----- PROTECTED REGION ID(OceanOptics::Additional Classes Definitions) ENABLED START -----*/

//	Additional Classes Definitions

/*----- PROTECTED REGION END -----*/	//	OceanOptics::Additional Classes Definitions

}	//	End of namespace

#endif   //	OceanOptics_H
