//=============================================================================
//
// file :        tcheck_get_rt_buffer.hpp
//
// description : Include with utility to read rt buffers.
//
// project :	RT BUFFER
//
// $Author: graziano $
//
// $Revision: 1.1 $
//
// $Log: tcheck_get_rt_buffer.hpp,v $
// Revision 1.1  2015-09-08 09:00:52  graziano
// upgraded to template buffer
//
//
//
//
// copyleft :   Sincrotrone Trieste S.C.p.A.
//              Strada Statale 14 - km 163,5 in AREA Science Park
//              34012 Basovizza (TS)
//              ITALY
//
//=============================================================================
#ifndef _T_CHECK_GET_RT_BUFFER_H
#define _T_CHECK_GET_RT_BUFFER_H

#include <tango.h>
#include <tbuffer.hpp>

namespace tbuffer
{

//+------------------------------------------------------------------
/**
 *	method:	check_get_data_param
 *
 *	description:	get data indexes
 *
 * @param[in]	buffer search params
 * @param[in]	rt buffer
 * @param[out]	buffer indexes
 * @param[out]	num samples between indexes
 * @param[in]	unique
 * @return	
 *
 */
//+------------------------------------------------------------------
template <typename T> void tcheck_get_data_param(const Tango::DevVarLongArray *argin, buffer_t<T> *tbuf, long *buf_idx, long *data_size, unsigned char *unique)
{
	long nargin = argin->length();
	long ret_num_samples;
	//struct _buffer *locbuf;
	long mode = (*argin)[0];
	 
	if ((*argin)[0] <= BUF_GET_TIME_MODE)
		*unique = 0;
	else if((*argin)[0] <= BUF_GET_TIME_UNIQUE_MODE)
		*unique = 1;
	else 
		*unique = 2;
	/* non unique modes (0,1,2); unique modes (3,4,5); bunch modes (6,7,8) */
	mode = mode - (*unique) * 3;
	
	/* get last values */
	if (nargin == 2)
	{
		long num_samples = (*argin)[1];

		if ((num_samples < 0) || (num_samples > tbuf->get_size()))
		{
			Tango::Except::throw_exception (
				(const char *)"Invalid number of data requested",
				(const char *)"Failed to retreive last data",
				(const char *)__FUNCTION__);
		}

		if (mode == BUF_GET_LAST_MODE)
		{
			ret_num_samples = tbuf->get_last_idx(num_samples, *unique, buf_idx);
		}
		else
		{
			Tango::Except::throw_exception (
				(const char *)"Wrong get mode parameter (0,1,2)",
				(const char *)"Failed to retreive last data",
				(const char *)__FUNCTION__);
		}

		if (ret_num_samples != num_samples && (*unique != 2))
		{
			stringstream tmp;
			tmp << " Can't retreive all data requested: requested="<<num_samples << " received="<<ret_num_samples;
			//DEBUG_STREAM << __FUNCTION__ << tmp.str();
			Tango::Except::throw_exception (
				(const char *)"Can't retreive all data requested",
				tmp.str(),
				(const char *)__FUNCTION__);
		}
		else if (ret_num_samples == 0)
		{
			stringstream tmp;
			tmp << " Can't retreive all data requested: requested="<<num_samples << " but buffer empty!";
			//DEBUG_STREAM << __FUNCTION__ << tmp.str();
			Tango::Except::throw_exception (
				(const char *)"Can't retreive all data requested",
				tmp.str(),
				(const char *)__FUNCTION__);
		}
		*data_size = ret_num_samples;
		if(*unique == 2)
			*data_size *= 2;	//concatenate bunchnumbers
	}
	/* get bunch mode values */
	else if (nargin == 3)
	{
		long bunch_start = (*argin)[1], bunch_end = (*argin)[2];

		if (((bunch_end-bunch_start) < 0) || ((bunch_end-bunch_start) > tbuf->get_size()))
		{
			stringstream out_stream;
			out_stream << "Invalid bunch numbers: bunch start=" << bunch_start << ", bunch end=" << bunch_end << ends;
			//DEBUG_STREAM << __FUNCTION__ << out_stream.str();
			Tango::Except::throw_exception (
				out_stream.str(),
				(const char *)"Failed to retreive bunch data",
				(const char *)__FUNCTION__);
		}

		if (mode == BUF_GET_BUNCH_MODE)
		{
			ret_num_samples = tbuf->get_bunch_idx(bunch_start, bunch_end, *unique, buf_idx);
		}
		else
		{
			stringstream out_stream;
			out_stream << "Invalid mode=" << mode << ends;
			//DEBUG_STREAM << __FUNCTION__ << out_stream.str();
			Tango::Except::throw_exception (
				out_stream.str(),
				(const char *)"Failed to retrive bunch data",
				(const char *)__FUNCTION__);
		}

		/* check data size */
		if ((((ret_num_samples-1) != (bunch_end-bunch_start)) && (*unique ==1)) ||
		    (((ret_num_samples-1) < (bunch_end-bunch_start)) && (*unique == 0)))
		{
			stringstream out_stream;
			out_stream << "Invalid bunch numbers: bunch start=" << bunch_start << ", bunch end=" << bunch_end <<  ", ret_num_samples=" <<
			(ret_num_samples-1) << ends;
			//DEBUG_STREAM << __FUNCTION__ << out_stream.str();
			Tango::Except::throw_exception (
				(const char *)"Failed to retreive bunch data",
				out_stream.str(),
				(const char *)__FUNCTION__);
		}
		else if(ret_num_samples == 0)
		{
			stringstream out_stream;
			out_stream << "Invalid bunch numbers.";
			if(buf_idx[0] == -1)
				out_stream<< " Bunch start=" << bunch_start << " not valid.";
			if(buf_idx[1] == -1)
				out_stream<< " Bunch end=" << bunch_end << " not valid.";
			out_stream << ends;
			//DEBUG_STREAM << __FUNCTION__ << out_stream.str();
			Tango::Except::throw_exception (
				(const char *)"Failed to retreive bunch data",
				out_stream.str(),
				(const char *)__FUNCTION__);
		}
		*data_size = ret_num_samples;
		if(*unique == 2)
			*data_size *= 2;	//concatenate bunchnumbers
	}
	/* get time mode values */
	else if (nargin == 5)
	{
		struct timeval time_start, time_end;

		time_start.tv_sec = (*argin)[1]; time_start.tv_usec = (*argin)[2];
		time_end.tv_sec = (*argin)[3]; time_end.tv_usec = (*argin)[4];

		if (((time_end.tv_sec-time_start.tv_sec)*1000000+(time_end.tv_usec-time_start.tv_usec)) < 0)
		{
			Tango::Except::throw_exception (
				(const char *)"Invalid time slice requested",
				(const char *)"Failed to retreive timed data",
				(const char *)__FUNCTION__);
		}

		if (mode == BUF_GET_TIME_MODE)
		{
			ret_num_samples = tbuf->get_time_idx(time_start, time_end, *unique, buf_idx);
		}
		else
		{
			Tango::Except::throw_exception (
				(const char *)"Wrong get mode parameter (0,1,2)",
				(const char *)"Failed to retreive timed data",
				(const char *)__FUNCTION__);
		}
		*data_size = ret_num_samples;
		if(*unique == 2)
			*data_size *= 2;	//concatenate bunchnumbers
	}
	/* invalid number of input parameters */
	else
	{
		Tango::Except::throw_exception (
			(const char *)"Invalid number of input parameters",
			(const char *)"Failed to get data",
			(const char *)__FUNCTION__);
	}	

}

template <typename T> void tcheck_get_data_param(const Tango::DevVarLongArray *argin, buffer_t<tbuffer::_data_array<T> > *tbuf, long *buf_idx, long *data_size, unsigned char *unique)
{
	long nargin = argin->length();
	long ret_num_samples;
	//struct _buffer *locbuf;
	long mode = (*argin)[0];

	if ((*argin)[0] <= BUF_GET_TIME_MODE)
		*unique = 0;
	else if((*argin)[0] <= BUF_GET_TIME_UNIQUE_MODE)
		*unique = 1;
	else
		*unique = 2;
	/* non unique modes (0,1,2); unique modes (3,4,5); bunch modes (6,7,8) */
	mode = mode - (*unique) * 3;

	/* get last values */
	if (nargin == 2)
	{
		long num_samples = (*argin)[1];

		if ((num_samples < 0) || (num_samples > tbuf->get_size()))
		{
			Tango::Except::throw_exception (
				(const char *)"Invalid number of data requested",
				(const char *)"Failed to retreive last data",
				(const char *)__FUNCTION__);
		}

		if (mode == BUF_GET_LAST_MODE)
		{
			ret_num_samples = tbuf->get_last_idx(num_samples, *unique, buf_idx);
		}
		else
		{
			Tango::Except::throw_exception (
				(const char *)"Wrong get mode parameter (0,1,2)",
				(const char *)"Failed to retreive last data",
				(const char *)__FUNCTION__);
		}

		if (ret_num_samples != num_samples && (*unique != 2))
		{
			stringstream tmp;
			tmp << " Can't retreive all data requested: requested="<<num_samples << " received="<<ret_num_samples;
			//DEBUG_STREAM << __FUNCTION__ << tmp.str();
			Tango::Except::throw_exception (
				(const char *)"Can't retreive all data requested",
				tmp.str(),
				(const char *)__FUNCTION__);
		}
		else if (ret_num_samples == 0)
		{
			stringstream tmp;
			tmp << " Can't retreive all data requested: requested="<<num_samples << " but buffer empty!";
			//DEBUG_STREAM << __FUNCTION__ << tmp.str();
			Tango::Except::throw_exception (
				(const char *)"Can't retreive all data requested",
				tmp.str(),
				(const char *)__FUNCTION__);
		}
		*data_size = ret_num_samples;
		if(*unique == 2)
			*data_size *= 2;	//concatenate bunchnumbers
	}
	/* get bunch mode values */
	else if (nargin == 3)
	{
		long bunch_start = (*argin)[1], bunch_end = (*argin)[2];

		if (((bunch_end-bunch_start) < 0) || ((bunch_end-bunch_start) > tbuf->get_size()))
		{
			stringstream out_stream;
			out_stream << "Invalid bunch numbers: bunch start=" << bunch_start << ", bunch end=" << bunch_end << ends;
			//DEBUG_STREAM << __FUNCTION__ << out_stream.str();
			Tango::Except::throw_exception (
				out_stream.str(),
				(const char *)"Failed to retreive bunch data",
				(const char *)__FUNCTION__);
		}

		if (mode == BUF_GET_BUNCH_MODE)
		{
			ret_num_samples = tbuf->get_bunch_idx(bunch_start, bunch_end, *unique, buf_idx);
		}
		else
		{
			stringstream out_stream;
			out_stream << "Invalid mode=" << mode << ends;
			//DEBUG_STREAM << __FUNCTION__ << out_stream.str();
			Tango::Except::throw_exception (
				out_stream.str(),
				(const char *)"Failed to retrive bunch data",
				(const char *)__FUNCTION__);
		}

		/* check data size */
		if ((((ret_num_samples-1) != (bunch_end-bunch_start)) && (*unique ==1)) ||
		    (((ret_num_samples-1) < (bunch_end-bunch_start)) && (*unique == 0)))
		{
			stringstream out_stream;
			out_stream << "Invalid bunch numbers: bunch start=" << bunch_start << ", bunch end=" << bunch_end <<  ", ret_num_samples=" <<
			(ret_num_samples-1) << ends;
			//DEBUG_STREAM << __FUNCTION__ << out_stream.str();
			Tango::Except::throw_exception (
				(const char *)"Failed to retreive bunch data",
				out_stream.str(),
				(const char *)__FUNCTION__);
		}
		else if(ret_num_samples == 0)
		{
			stringstream out_stream;
			out_stream << "Invalid bunch numbers.";
			if(buf_idx[0] == -1)
				out_stream<< " Bunch start=" << bunch_start << " not valid.";
			if(buf_idx[1] == -1)
				out_stream<< " Bunch end=" << bunch_end << " not valid.";
			out_stream << ends;
			//DEBUG_STREAM << __FUNCTION__ << out_stream.str();
			Tango::Except::throw_exception (
				(const char *)"Failed to retreive bunch data",
				out_stream.str(),
				(const char *)__FUNCTION__);
		}
		*data_size = ret_num_samples;
		if(*unique == 2)
			*data_size *= 2;	//concatenate bunchnumbers
	}
	/* get time mode values */
	else if (nargin == 5)
	{
		struct timeval time_start, time_end;

		time_start.tv_sec = (*argin)[1]; time_start.tv_usec = (*argin)[2];
		time_end.tv_sec = (*argin)[3]; time_end.tv_usec = (*argin)[4];

		if (((time_end.tv_sec-time_start.tv_sec)*1000000+(time_end.tv_usec-time_start.tv_usec)) < 0)
		{
			Tango::Except::throw_exception (
				(const char *)"Invalid time slice requested",
				(const char *)"Failed to retreive timed data",
				(const char *)__FUNCTION__);
		}

		if (mode == BUF_GET_TIME_MODE)
		{
			ret_num_samples = tbuf->get_time_idx(time_start, time_end, *unique, buf_idx);
		}
		else
		{
			Tango::Except::throw_exception (
				(const char *)"Wrong get mode parameter (0,1,2)",
				(const char *)"Failed to retreive timed data",
				(const char *)__FUNCTION__);
		}
		*data_size = ret_num_samples;
		if(*unique == 2)
			*data_size *= 2;	//concatenate bunchnumbers
	}
	/* invalid number of input parameters */
	else
	{
		Tango::Except::throw_exception (
			(const char *)"Invalid number of input parameters",
			(const char *)"Failed to get data",
			(const char *)__FUNCTION__);
	}

}

}//namespace
#endif		//_T_CHECK_GET_RT_BUFFER_H
